from langchain_experimental.tools.python.tool import PythonAstREPLTool
from .agent import BaseAgent
# importing Tools
from ..tools import (
    # ZipcodeAPIToolkit,
    # WikipediaTool,
    # WikidataTool,
    GoogleSearchTool,
    GoogleLocationFinder,
    BingSearchTool,
    # AskNewsTool,
    DuckDuckGoSearchTool,
    YouTubeSearchTool,
    OpenWeatherMapTool,
    StackExchangeTool,
)
from ..tools.execute import ExecutablePythonREPLTool

# ZipCode API Toolkit
# zpt = ZipcodeAPIToolkit()
# zpt_tools = zpt.get_tools()

# wk1 = WikipediaTool()
# wk12 = WikidataTool()

g1 = GoogleSearchTool()
g2 = GoogleLocationFinder()

b = BingSearchTool()
d = DuckDuckGoSearchTool()
# ask = AskNewsTool()

yt = YouTubeSearchTool()
stackexchange = StackExchangeTool()
weather = OpenWeatherMapTool()

tooling = [
    # wk1,
    g1, g2,
    b, d, yt,
    weather,
    stackexchange
] # + zpt_tools

class CopilotAgent(BaseAgent):
    """CopilotAgent Agent.

    This is Agent Base class for AI Copilots.
    """
    def __init__(
        self,
        name: str = 'Agent',
        llm: str = 'vertexai',
        tools: list = None,
        prompt_template: str = None,
        **kwargs
    ):
        super().__init__(name, llm, tools, prompt_template, **kwargs)
        if not tools:
            tools = tooling
        self.tools = [
                PythonAstREPLTool(
                    name='python_repl_ast',
                    globals={},
                    locals={}
                ),
                ExecutablePythonREPLTool(
                    name='executable_python_repl_ast',
                    globals={},
                    locals={}
                )
            ] + list(tools)
        self.prompt = self.get_prompt(
            self.prompt_template
        )
        print('PROMPT > ', self.prompt)

    @classmethod
    def default_tools(cls) -> list:
        # ZipCode API Toolkit
        tools = []
        zpt_tools = []
        try:
            zpt = ZipcodeAPIToolkit()
            zpt_tools = zpt.get_tools()

        except Exception as e:
            print('ERROR LOADING ZIPCODE TOOLS')

        try:
            # wk1 = WikipediaTool()
            # wk12 = WikidataTool()

            g1 = GoogleSearchTool()
            g2 = GoogleLocationFinder()

            b = BingSearchTool()
            d = DuckDuckGoSearchTool()
            # ask = AskNewsTool()

            yt = YouTubeSearchTool()
            stackexchange = StackExchangeTool()
            weather = OpenWeatherMapTool()

            tools = [
                # wk1,wk12,
                g1, g2,
                b, d, # ask,
                yt,
                weather,
                stackexchange
            ]

        except Exception as e:
            print('TOOL Error > ', e)

        return tools



import torch
from langchain_community.llms import HuggingFacePipeline  # pylint: disable=import-error, E0611
from transformers import (
    AutoModelForCausalLM,
    AutoProcessor,
    LlavaForConditionalGeneration,
    AutoTokenizer,
    GenerationConfig,
    pipeline
)
from .abstract import AbstractLLM


class PipelineLLM(AbstractLLM):
    """PipelineLLM.

    Load a LLM (Language Model) from HuggingFace Hub.

    Returns:
        _type_: an instance of HuggingFace LLM Model.
    """
    model: str = "databricks/dolly-v2-3b"
    embed_model: str = None
    max_tokens: int = 1024
    supported_models: list = [
        "databricks/dolly-v2-3b",
        "gpt2",
        "bigscience/bloom-1b7",
        "meta-llama/Llama-2-7b-hf",
        'llava-hf/llava-1.5-7b-hf'
    ]

    def __init__(self, *args, **kwargs):
        self.batch_size = kwargs.get('batch_size', 4)
        self.use_llava: bool = kwargs.get('use_llava', False)
        self.model_args = kwargs.get('model_args', {})
        super().__init__(*args, **kwargs)
        dtype = kwargs.get('dtype', 'float16')
        if dtype == 'bfloat16':
            torch_dtype = torch.bfloat16
        if dtype == 'float16':
            torch_dtype = torch.float16
        elif dtype == 'float32':
            torch_dtype = torch.float32
        elif dtype == 'float8':
            torch_dtype = torch.float8
        else:
            torch_dtype = "auto"
        use_fast = kwargs.get('use_fast', True)
        if self.use_llava is False:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.model,
                chunk_size=self.max_tokens
            )
            self._model = AutoModelForCausalLM.from_pretrained(
                self.model,
                device_map="auto",
                torch_dtype=torch_dtype,
                trust_remote_code=True,
            )
            config = GenerationConfig(
                do_sample=True,
                temperature=self.temperature,
                max_new_tokens=self.max_tokens,
                top_p=self.top_p,
                top_k=self.top_k,
                repetition_penalty=1.15,
            )
            self._pipe = pipeline(
                task=self.task,
                model=self._model,
                tokenizer=self.tokenizer,
                return_full_text=True,
                use_fast=use_fast,
                device_map='auto',
                batch_size=self.batch_size,
                generation_config=config,
                pad_token_id = 50256,
                framework="pt"
            )
        else:
            self._model = LlavaForConditionalGeneration.from_pretrained(
                self.model,
                device_map="auto",
                torch_dtype=torch_dtype,
                trust_remote_code=True,
                low_cpu_mem_usage=True,
            )
            self.tokenizer = AutoTokenizer.from_pretrained(self.model)
            processor = AutoProcessor.from_pretrained(self.model)
            self._pipe = pipeline(
                task=self.task,
                model=self._model,
                tokenizer=self.tokenizer,
                use_fast=use_fast,
                device_map='auto',
                batch_size=self.batch_size,
                image_processor=processor.image_processor,
                framework="pt",
                **self.model_args
            )
        self._pipe.tokenizer.pad_token_id = self._pipe.model.config.eos_token_id
        self._llm = HuggingFacePipeline(
            model_id=self.model,
            pipeline=self._pipe,
            verbose=True
        )

    def pipe(self, *args, **kwargs):
        return self._pipe(
            *args,
            **kwargs,
            generate_kwargs={"max_new_tokens": self.max_tokens}
        )



from langchain_core.retrievers import BaseRetriever
from langchain_core.vectorstores import VectorStoreRetriever


class EmptyRetriever(BaseRetriever):
    """Return a Retriever with No results.
    """
    async def aget_relevant_documents(self, query: str):
        return []

    def _get_relevant_documents(self, query: str):
        return []



from .chatbot import Chatbot

class AskTROC(Chatbot):
    """Represents an Human Resources agent in Navigator.

        Each agent has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    company_information: dict = {
        'company': 'T-ROC Global',
        'company_website': 'https://www.trocglobal.com',
        'contact_email': 'communications@trocglobal.com',
        'contact_form': 'https://www.surveymonkey.com/r/TROC_Suggestion_Box'
    }
    role: str = 'Expert T-ROCer'
    goal = 'Bring useful information about T-ROC Global to employees.'



from typing import List, Dict, Any, Optional, Type, Union
import httpx
from pydantic import BaseModel, Field
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.tools import BaseToolkit
from langchain_core.tools import BaseTool
from navconfig import config
from .abstract import AbstractTool


class ZipcodeDistanceInput(BaseModel):
    """Input for the Zipcode Distance Tool."""
    zipcode1: Union[str, int] = Field(description="The first zipcode.")
    zipcode2: Union[str, int] = Field(description="The second zipcode.")
    unit: Optional[str] = Field(description="The unit of the distance.", default="mile")

    model_config = {
        "arbitrary_types_allowed": True,
        "extra": "forbid",  # Helps with compatibility
        "json_schema_extra": {
            "required": ["zipcode", "sku", "location_id"]
        }
    }

class ZipcodeRadiusInput(BaseModel):
    """Input for the Zipcode Radius Tool."""
    zipcode: Union[str, int] = Field(description="The zipcode.")
    radius: int = Field(description="The radius in miles.", default=5)
    unit: Optional[str] = Field(description="The unit of the distance.", default="mile")

    model_config = {
        "arbitrary_types_allowed": True,
        "extra": "forbid",  # Helps with compatibility
        "json_schema_extra": {
            "required": ["zipcode", "sku", "location_id"]
        }
    }

class ZipcodeLocationInput(BaseModel):
    """Input for the Zipcode Location Tool."""
    zipcode: Union[str, int] = Field(description="The zipcode.")
    unit: Optional[str] = Field(description="The unit of the distance.", default="degrees")

    model_config = {
        "arbitrary_types_allowed": True,
        "extra": "forbid",  # Helps with compatibility
        "json_schema_extra": {
            "required": ["zipcode", "sku", "location_id"]
        }
    }

class ZipcodeDistance(AbstractTool):
    """Tool for calculating the distance between two zipcodes."""

    name: str = "zipcode_distance"
    verbose: bool = True
    args_schema: Type[BaseModel] = ZipcodeDistanceInput
    description: str = (
        "Use this tool to calculate the distance between two zipcodes."
        " Zipcodes must be provided as a couple of strings (e.g., '33066')."
    )

    model_config = {
        "arbitrary_types_allowed": True,
        "extra": "forbid",  # Helps with compatibility
        "json_schema_extra": {
            "required": ["zipcode", "sku", "location_id"]
        }
    }

    def _search(
        self,
        zipcode1: str,
        zipcode2: str,
        unit: Optional[str] = 'mile',
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Dict[str, Any]:  # Changed to Dict
        api_key = config.get('ZIPCODE_API_KEY')
        url = f"https://www.zipcodeapi.com/rest/{api_key}/distance.json/{zipcode1}/{zipcode2}/{unit}"

        try:
            response = httpx.get(url)
            response.raise_for_status()  # Check for API errors
            return response.json()
        except httpx.HTTPStatusError as e:
            raise ValueError(f"Error fetching zipcode distance: {e}") from e

    async def _asearch(
        self,
        zipcode1: str,
        zipcode2: str,
        unit: Optional[str] = 'mile',
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Dict[str, Any]:
        api_key = config.get('ZIPCODE_API_KEY')
        url = f"https://www.zipcodeapi.com/rest/{api_key}/distance.json/{zipcode1}/{zipcode2}/{unit}"

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(url)
                response.raise_for_status()  # Check for API errors
                return response.json()
        except httpx.HTTPStatusError as e:
            raise ValueError(f"Error fetching zipcode distance: {e}") from e


class ZipcodeRadius(AbstractTool):
    """Tool for calculating the distance between two zipcodes."""

    name: str = "zipcodes_by_radius"
    verbose: bool = True
    args_schema: Type[BaseModel] = ZipcodeRadiusInput
    description: str = (
        "Use this Tool to find all US zip codes within a given radius of a zip code."
        " Provides a Zipcode and a radius."
    )

    model_config = {
        "arbitrary_types_allowed": True,
        "extra": "forbid",  # Helps with compatibility
        "json_schema_extra": {
            "required": ["zipcode", "sku", "location_id"]
        }
    }

    def _search(
        self,
        zipcode: Union[str, int],
        radius: int = 5,
        unit: Optional[str] = 'mile',
    ) -> Dict[str, Any]:  # Changed to Dict
        api_key = config.get('ZIPCODE_API_KEY')
        url = f"https://www.zipcodeapi.com/rest/{api_key}/radius.json/{zipcode}/{radius}/{unit}"
        try:
            response = httpx.get(url)
            response.raise_for_status()  # Check for API errors
            return response.json()
        except httpx.HTTPStatusError as e:
            raise ValueError(f"Error fetching zipcode distance: {e}") from e

    async def _asearch(
        self,
        zipcode: str,
        radius: int,
        unit: Optional[str] = 'mile'
    ) -> Dict[str, Any]:
        api_key = config.get('ZIPCODE_API_KEY')
        url = f"https://www.zipcodeapi.com/rest/{api_key}/radius.json/{zipcode}/{radius}/{unit}"

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(url)
                response.raise_for_status()  # Check for API errors
                return response.json()
        except httpx.HTTPStatusError as e:
            raise ValueError(f"Error fetching zipcode distance: {e}") from e



class ZipcodeLocation(AbstractTool):
    """Tool for calculating Geographical information about a Zipcode."""

    name: str = "zipcode_location"
    verbose: bool = True
    args_schema: Type[BaseModel] = ZipcodeLocationInput
    description: str = (
        "Use this Tool to find out the city, state, latitude, longitude, and time zone information for a US zip code."
        " Use this tool to find geographical information about a zipcode. "
        " Provides only a Zipcode as string."
    )

    def _search(
        self,
        zipcode: str,
        unit: Optional[str] = 'degrees'
    ) -> Dict[str, Any]:  # Changed to Dict
        api_key = config.get('ZIPCODE_API_KEY')
        url = f"https://www.zipcodeapi.com/rest/{api_key}/info.json/{zipcode}/{unit}"

        try:
            response = httpx.get(url)
            response.raise_for_status()  # Check for API errors
            return response.json()
        except httpx.HTTPStatusError as e:
            raise ValueError(f"Error fetching zipcode distance: {e}") from e


class ZipcodeAPIToolkit(BaseToolkit):
    """Toolkit for interacting with ZipcodeAPI.
    """

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
            ZipcodeLocation(),
            ZipcodeRadius(),
            ZipcodeDistance()
        ]



from langchain.tools import Tool
from langchain.tools import BaseTool
from langchain_community.utilities import StackExchangeAPIWrapper

class StackExchangeTool(BaseTool):
    """Tool that searches the StackExchangeTool API."""
    name: str = "StackExchangeSearch"
    description: str = (
        "Search for questions and answers on Stack Exchange. "
        "Stack Exchange is a network of question-and-answer (Q&A) websites on topics in diverse fields, each site covering a specific topic."
        "Useful for when you need to answer general questions about different topics when user requested."
    )
    search: Tool = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.search = StackExchangeAPIWrapper(
            query_type='title',
            max_results=5
        )

    def _run(
        self,
        query: dict,
    ) -> dict:
        """Use the StackExchangeSearch tool."""
        return self.search.run(query)



from typing import Union
import pandas as pd
from langchain_experimental.tools.python.tool import PythonAstREPLTool
from .base import BaseAgent


BASEPROMPT = """
Your name is {name}, You are a helpful assistant built to provide comprehensive guidance and support on data calculations and various aspects of payroll, such as earnings, deductions, salary information, and payroll tax. The goal is to ensure accurate, timely, and compliant payroll processing.

**Your Capabilities:**

1. **Earnings Calculation:**
   - Calculate regular earnings based on hourly rates or salaries.
   - Handle overtime calculations, including differentials and double-time as needed.
   - Include bonuses, commissions, and other incentive earnings.

2. **Deductions:**
   - Calculate mandatory deductions including federal, state, and local taxes.
   - Handle other withholdings such as Social Security, Medicare, disability insurance, and unemployment insurance.
   - Process voluntary deductions (e.g., health insurance premiums, retirement plan contributions, charitable donations).

3. **Salary Information:**
   - Provide gross and net salary breakdowns.
   - Assist in setting up and adjusting salary structures.
   - Manage payroll for part-time, full-time, and contract employees.

These keywords must never be translated and transformed:
- Action:
- Thought:
- Action Input:
because they are part of the thinking process instead of the output.

You are working with {num_dfs} pandas dataframes in Python named df1, df2, etc. You should use the tools below to answer the question posed to you:

- python_repl_ast: A Python shell. Use this to execute python commands. Input should be a valid python command. When using this tool, sometimes output is abbreviated - make sure it does not look abbreviated before using it in your answer.
- {tools}

To use a tool, please use the following format:

```
Question: the input question you must answer.
Thought: You should always think about what to do, don't try to reason out the answer on your own.
Action: the action to take, should be one of [python_repl_ast,{tool_names}] if using a tool, otherwise answer on your own.
Action Input: the input to the action.
Observation: the result of the action.
... (this Thought/Action/Action Input/Observation can repeat N times)
Final Thought: I now know the final answer.
Final Answer: the final answer to the original input question.
```


```
This is the result of `print(df1.head())` for each dataframe:
{dfs_head}
```

Begin!

Question: {input}
{agent_scratchpad}
"""


class DataFrameAgent(BaseAgent):
    """Dataframe Agent.

    This is the Pandas Agent Chatbot.
    """
    def __init__(
        self,
        name: str = 'Agent',
        llm: str = 'vertexai',
        tools: list = None,
        df: Union[list[pd.DataFrame], pd.DataFrame] = None,
        prompt_template: str = None,
        **kwargs
    ):
        super().__init__(name, llm, tools, prompt_template, **kwargs)
        for _df in df if isinstance(df, list) else [df]:
            if not isinstance(_df, pd.DataFrame):
                raise ValueError(
                    f"Expected pandas DataFrame, got {type(_df)}"
                )
        self.df = df
        df_locals = {}
        dfs_head = ""
        num_dfs = 1
        if isinstance(df, pd.DataFrame):
            df = [df]
        num_dfs = len(df)
        for i, dataframe in enumerate(df):
                    df_locals[f"df{i + 1}"] = dataframe
                    dfs_head += (
                        f"\n\n- This is the result of `print(df{i + 1}.head())`:\n"
                        + dataframe.head().to_markdown() + "\n"
                    )
        prompt_template = BASEPROMPT
        self.tools = [PythonAstREPLTool(locals=df_locals)] + list(tools)
        self.prompt = self.get_prompt(
            prompt_template,
            num_dfs=num_dfs,
            dfs_head=dfs_head
        )



from langchain_community.embeddings import HuggingFaceBgeEmbeddings
from .base import BaseEmbed


class BgeEmbed(BaseEmbed):
    """A wrapper class for VertexAI embeddings."""
    model_name: str = "BAAI/bge-large-en-v1.5"

    def _create_embedding(self, model_name: str = None, **kwargs):
        # Embedding Model:
        device = self._get_device()
        model_args = {
            **self.model_kwargs,
            'device': device,
        }
        return HuggingFaceBgeEmbeddings(
            model_name=model_name or self.model_name,
            model_kwargs=model_args,
            encode_kwargs=self.encode_kwargs
        )



from collections.abc import Callable
from typing import List, Optional, Union
from uuid import uuid4
from langchain.docstore.document import Document
# Milvus Database
from pymilvus import (
    db,
    connections,
    FieldSchema,
    CollectionSchema,
    DataType,
    Collection,
    MilvusClient
)
from pymilvus.exceptions import MilvusException
from langchain_milvus import Milvus  # pylint: disable=import-error, E0611
from langchain.memory import VectorStoreRetrieverMemory
from navconfig.logging import logging
from .abstract import AbstractStore
from ..conf import (
    MILVUS_HOST,
    MILVUS_PROTOCOL,
    MILVUS_PORT,
    MILVUS_URL,
    MILVUS_TOKEN,
    MILVUS_USER,
    MILVUS_PASSWORD,
    MILVUS_SECURE,
    MILVUS_SERVER_NAME,
    MILVUS_CA_CERT,
    MILVUS_SERVER_CERT,
    MILVUS_SERVER_KEY,
    MILVUS_USE_TLSv2
)


logging.getLogger(name='pymilvus').setLevel(logging.WARNING)

class MilvusConnection:
    """
    Context Manager for Milvus Connections.
    """
    def __init__(self, alias: str = 'default', **kwargs):
        self._connected: bool = False
        self.credentials = kwargs
        self.alias: str = alias

    def connect(self, alias: str = None, **kwargs):
        if not alias:
            alias = self.alias
        conn = connections.connect(
            alias=alias,
            **kwargs
        )
        self._connected = True
        return alias

    def is_connected(self):
        return self._connected

    def close(self, alias: str = None):
        try:
            connections.disconnect(alias=alias)
        finally:
            self._connected = False

    def __enter__(self):
        self.connect(alias=self.alias, **self.credentials)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close(alias=self.alias)
        return self


class MilvusStore(AbstractStore):
    """MilvusStore class.

    Milvus is a Vector Database multi-layered.

    Args:
        host (str): Milvus host.
        port (int): Milvus port.
        url (str): Milvus URL.
    """

    def __init__(
        self,
        embedding_model: Union[dict, str] = None,
        embedding: Union[dict, Callable] = None,
        **kwargs
    ):
        self.host = kwargs.pop("host", MILVUS_HOST)
        self.port = kwargs.pop("port", MILVUS_PORT)
        self.protocol = kwargs.pop("protocol", MILVUS_PROTOCOL)
        self.consistency_level: str = kwargs.pop('consistency_level', 'Session')
        self.create_database: bool = kwargs.pop('create_database', True)
        self.url = kwargs.pop("url", MILVUS_URL)
        self._client_id = kwargs.pop('client_id', 'default')
        super().__init__(embedding_model, embedding, **kwargs)
        if not self.url:
            self.url = f"{self.protocol}://{self.host}:{self.port}"
        else:
            # Extract host and port from URL
            if not self.host:
                self.host = self.url.split("://")[-1].split(":")[0]
            if not self.port:
                self.port = int(self.url.split(":")[-1])
        self.token = kwargs.pop("token", MILVUS_TOKEN)
        # user and password (if required)
        self.user = kwargs.pop(
            "user", MILVUS_USER
        )
        self.password = kwargs.pop(
            "password", MILVUS_PASSWORD
        )
        # SSL/TLS
        self._secure: bool = kwargs.pop('secure', MILVUS_SECURE)
        self._server_name: str = kwargs.pop('server_name', MILVUS_SERVER_NAME)
        self._cert: str = kwargs.pop('server_pem_path', MILVUS_SERVER_CERT)
        self._ca_cert: str = kwargs.pop('ca_pem_path', MILVUS_CA_CERT)
        self._cert_key: str = kwargs.pop('client_key_path', MILVUS_SERVER_KEY)
        # Any other argument will be passed to the Milvus client
        self.credentials = {
            "uri": self.url,
            "host": self.host,
            "port": self.port,
            **kwargs
        }
        if self.token:
            self.credentials['token'] = self.token
        if self.user:
            self.credentials['token'] = f"{self.user}:{self.password}"
        # SSL Security:
        if self._secure is True:
            args = {
                "secure": self._secure,
                "server_name": self._server_name
            }
            if self._cert:
                if MILVUS_USE_TLSv2 is True:
                    args['client_pem_path'] = self._cert
                    args['client_key_path'] = self._cert_key
                else:
                    args["server_pem_path"] = self._cert
            if self._ca_cert:
                args['ca_pem_path'] = self._ca_cert
            self.credentials = {**self.credentials, **args}
        if self.database:
            self.credentials['db_name'] = self.database

    async def connection(self, alias: str = None) -> "MilvusStore":
        """Connects to the Milvus database."""
        self._client_id = alias or uuid4().hex
        _ = connections.connect(
            alias=self._client_id,
            **self.credentials
        )
        try:
            if self.database:
                self.use_database(
                    self.database,
                    alias=self._client_id,
                    create=self.create_database
                )
        except Exception as e:
            self.logger.error(
                f"Cannot create Database {self.database} for alias {self._client_id}: {e}"
            )
        self._connection = MilvusClient(
            **self.credentials
        )
        self._connected = True
        print('Connected to database', self._connection)
        return self

    async def disconnect(self, alias: str = None):
        try:
            a = alias or self._client_id
            connections.disconnect(alias=a)
            self._connection.close()
        except AttributeError:
            pass
        finally:
            self._connection = None
            self.client = None
            self._client_id = None
            self._connected = False

    def use_database(
        self,
        db_name: str,
        alias: str = 'default',
        create: bool = False
    ) -> None:
        try:
            conn = connections.connect(alias, **self.credentials)
        except MilvusException as exc:
            if "database not found" in exc.message:
                self.logger.error(
                    f"Database {db_name} does not exist."
                )
                args = self.credentials.copy()
                del args['db_name']
                db.create_database(db_name, alias=alias, **args)
        # re-connect:
        try:
            _ = connections.connect(alias, **self.credentials)
            if db_name not in db.list_database(using=alias):
                if self.create_database is True or create is True:
                    try:
                        db.create_database(db_name, using=alias, timeout=10)
                        self.logger.notice(
                            f"Database {db_name} created successfully."
                        )
                    except Exception as e:
                        raise ValueError(
                            f"Error creating database: {e}"
                        )
                else:
                    raise ValueError(
                        f"Database {db_name} does not exist."
                    )
        finally:
            connections.disconnect(alias=alias)

    def setup_vector(self):
        self.vector = Milvus(
            self._embed_,
            consistency_level='Bounded',
            connection_args={**self.credentials},
            collection_name=self.collection_name,
        )
        return self.vector

    def get_vectorstore(self):
        return self.get_vector()

    async def collection_exists(self, collection_name: str) -> bool:
        async with self.connection():
            if collection_name in self._connection.list_collections():
                return True
        return False

    def check_state(self, collection_name: str) -> dict:
        return self._connection.get_load_state(
            collection_name=collection_name
        )

    def get_vector(
        self,
        collection: Union[str, None] = None,
        metric_type: str = None,
        index_type: str = None,
        nprobe: int = 32,
        metadata_field: str = None,
        consistency_level: str = 'session'
    ) -> Milvus:
        if not metric_type:
            metric_type = self._metric_type
        if not collection:
            collection = self.collection_name
        if not metric_type:
            metric_type = self._metric_type or 'COSINE'
        _idx = index_type or self._index_type
        _search = {
            "search_params": {
                "metric_type": metric_type,
                "index_type": _idx,
                "params": {"nprobe": nprobe, "nlist": 1024},
            }
        }
        if metadata_field:
            # document_meta
            _search['metadata_field'] = metadata_field
        if self._embed_ is None:
            self._embed_ = self.create_embedding(
                embedding_model=self.embedding_model
            )
        return Milvus(
            embedding_function=self._embed_.embedding,
            collection_name=collection,
            consistency_level=consistency_level,
            connection_args={
                **self.credentials
            },
            primary_field='pk',
            text_field='text',
            vector_field='vector',
            **_search
        )

    def search(
        self,
        payload: Union[dict, list],
        collection: Union[str, None] = None,
        limit: Optional[int] = None
    ) -> list:
        args = {}
        if collection is None:
            collection = self.collection_name
        if limit is not None:
            args = {"limit": limit}
        if isinstance(payload, dict):
            payload = [payload]
        result = self._connection.search(
            collection_name=collection,
            data=payload,
            **args
        )
        return result

    def memory_retriever(self, documents: List[Document], num_results: int  = 5) -> VectorStoreRetrieverMemory:
        vectordb = Milvus.from_documents(
            documents or [],
            self._embed_,
            connection_args={**self.credentials}
        )
        retriever = Milvus.as_retriever(
            vectordb,
            search_kwargs=dict(k=num_results)
        )
        return VectorStoreRetrieverMemory(retriever=retriever)

    def save_context(self, memory: VectorStoreRetrieverMemory, context: list) -> None:
        for val in context:
            memory.save_context(val)

    async def similarity_search(
        self,
        query: str,
        collection: Union[str, None] = None,
        limit: int = 2,
        consistency_level: str = 'Bounded'
    ) -> list:
        if collection is None:
            collection = self.collection_name
        if self._embed_ is None:
            _embed_ = self.create_embedding(
                embedding_model=self.embedding_model
            )
        else:
            _embed_ = self._embed_
        vector_db = Milvus(
            embedding_function=_embed_,
            collection_name=collection,
            consistency_level=consistency_level,
            connection_args={
                **self.credentials
            },
            primary_field='pk',
            text_field='text',
            vector_field='vector'
        )
        return await vector_db.asimilarity_search(query, k=limit)

    async def insert(self, collection: str, embeddings: list) -> dict:
        if not collection:
            collection = self.collection_name
        async with self:
            # create the collection object from milvus
            if self._connection.has_collection(collection):
                schema = self._connection.describe_collection(collection)
                if schema.params['dimension']!= embeddings[0].shape[0]:
                    raise ValueError(
                        f"Invalid dimension: expected {schema.params['dimension']}, got {embeddings[0].shape[0]}"
                    )
                self._connection.upsert(collection, embeddings)
            else:
                raise RuntimeError(
                    f"Collection {collection} does not exist."
                )

    async def from_documents(self, documents: list[Document], collection: str = None, **kwargs):
        """
        Save Documents as Vectors in Milvus.
        """
        if not collection:
            collection = self.collection_name
        vectordb = await Milvus.afrom_documents(
            documents=documents,
            embedding=self._embed_,
            connection_args={**self.credentials},
            collection_name=collection
        )
        return vectordb

    async def add_documents(self, documents: list[Document], collection: str = None, **kwargs):
        """
        Add Documents as Vectors in Milvus.
        """
        if not collection:
            collection = self.collection_name
        async with self:
            # create the collection object from milvus
            vectordb = self.get_vector(collection=collection, **kwargs)
            await vectordb.aadd_documents(documents)




def agent_config(
    provider: str = 'vertexai',
    model: str = 'gemini-pro',
    temperature: float = 0.4,
    top_p: int = 0.5,
    stream: bool = True
    ) -> dict:
    agent_config ={
        # "llm": {
        #     "provider": provider,
        #     "config": {
        #         "model": model,
        #         "temperature": temperature,
        #         "top_p": top_p,
        #         "stream": stream
        #     }
        # },
        "embedder": {
            "provider": "huggingface",
            "config": {
                "model": "sentence-transformers/sentence-t5-large",
            }
        }
    }
    return agent_config


def agent_embedder(
    provider: str = 'huggingface',
    model: str = 'sentence-transformers/sentence-t5-large',
    **kwargs
    ) -> dict:
    agent_config ={
        "embedder": {
            "provider": "huggingface",
            "config": {
                "model": "sentence-transformers/sentence-t5-large",
                **kwargs
            }
        }
    }
    return agent_config



from collections.abc import Callable
from pathlib import PurePath
from typing import Optional, Union
from uuid import uuid4
import faiss
from langchain.docstore.document import Document
from langchain.memory import VectorStoreRetrieverMemory
from langchain_community.vectorstores import FAISS
from langchain_community.docstore.in_memory import InMemoryDocstore
from langchain_community.vectorstores.utils import DistanceStrategy
from .abstract import AbstractStore


class FaissStore(AbstractStore):
    """FAISS DB Store Class.

    Using FAISS as an in-memory Document Vector Store.

    """

    def __init__(
        self,
        embedding_model: Union[dict, str] = None,
        embedding: Union[dict, Callable] = None,
        **kwargs
    ):
        super().__init__(
            embedding_model=embedding_model,
            embedding=embedding,
            **kwargs
        )
        self.index_path: PurePath = kwargs.pop('index_path', 'faiss_index')
        if not self.index_path.exists():
            self.index_path.mkdir(exist_ok=True)

    async def connection(self):
        """Initialize FAISS vector store.

        If an index exists, load it; otherwise, create a new FAISS store.
        """
        try:
            index_file = self.index_path.joinpath("index.faiss")
            if not index_file.exists():
                raise FileNotFoundError
            self._connection = FAISS.load_local(
                folder_path=self.index_path,
                embeddings=self._embed_.embedding,
                allow_dangerous_deserialization=True,
                distance_strategy=DistanceStrategy.EUCLIDEAN_DISTANCE
            )
        except FileNotFoundError:
            # Create a new FAISS index if none exists.
            print(len(self._embed_.embedding.embed_query("test")))
            index = faiss.IndexFlatL2(len(self._embed_.embedding.embed_query("test")))
            self._connection = FAISS(
                embedding_function=self._embed_.embedding,
                index=index,
                docstore=InMemoryDocstore(),
                index_to_docstore_id={},
                normalize_L2=True,
                distance_strategy=DistanceStrategy.EUCLIDEAN_DISTANCE
            )
            self._connection.save_local(self.index_path)  # Save the new index
        self._connected = True
        return self._connection

    async def disconnect(self) -> None:
        """Clears FAISS in-memory index."""
        self._connection = None
        self._connected = False

    def get_vector(
        self,
        embedding: Optional[Callable] = None,
    ) -> FAISS:
        """Returns FAISS VectorStore instance."""
        if embedding is not None:
            _embed_ = embedding
        else:
            _embed_ = self.create_embedding(
                embedding_model=self.embedding_model
            )
        return FAISS.load_local(
            folder_path=self.index_path,
            embeddings=_embed_,
            allow_dangerous_deserialization=True,
            distance_strategy=DistanceStrategy.EUCLIDEAN_DISTANCE
        )

    async def from_documents(self, documents: list[Document], **kwargs):
        """Save Documents as Vectors in FAISS."""
        vectordb = await FAISS.afrom_documents(
            documents=documents,
            embedding=self._embed_.embedding,
            allow_dangerous_deserialization=True,
            distance_strategy=DistanceStrategy.EUCLIDEAN_DISTANCE
        )
        vectordb.save_local(self.index_path)  # Persist FAISS index
        return vectordb

    async def add_documents(
        self,
        documents: list,
        embedding: Optional[Callable] = None,
    ) -> bool:
        """Add Documents to FAISS."""
        async with self:
            vector_db = self.get_vector(embedding=embedding)
            await vector_db.aadd_documents(documents=documents)
            vector_db.save_local(self.index_path)  # Save updated index
        return True

    async def update_documents(
        self,
        documents: list,
        embedding: Optional[Callable] = None,
    ) -> bool:
        """
        Update Documents in FAISS (FAISS does not natively support updates).
        """
        async with self:
            vector_db = self.get_vector(embedding=embedding)
            if all('id' in doc for doc in documents):
                ids = [doc.pop('id') for doc in documents]
                vector_db.delete(ids)  # Remove old entries
                await vector_db.aadd_documents(documents=documents)  # Add new versions
                vector_db.save_local(self.index_path)
                return True
            return False

    async def similarity_search(
        self,
        query: str,
        embedding: Optional[Callable] = None,
        limit: int = 2,
    ) -> list:
        """Performs similarity search in FAISS."""
        async with self:
            vector_db = self.get_vector(embedding=embedding)
            return vector_db.similarity_search(query, k=limit)

    def memory_retriever(
        self,
        documents: Optional[list] = None,
        num_results: int = 5
    ) -> VectorStoreRetrieverMemory:
        """Retrieves stored memory-based documents."""
        if not documents:
            documents = []
        vectordb = FAISS.from_documents(
            documents=documents,
            embedding=self._embed_.embedding,
            allow_dangerous_deserialization=True,
            distance_strategy=DistanceStrategy.EUCLIDEAN_DISTANCE
        )
        retriever = FAISS.as_retriever(
            vectordb,
            search_kwargs=dict(k=num_results)
        )
        return VectorStoreRetrieverMemory(retriever=retriever)



from langchain_google_vertexai import VertexAIEmbeddings
from ...conf import VERTEX_PROJECT_ID, VERTEX_REGION
from .abstract import AbstractEmbed

class VertexAIEmbed(AbstractEmbed):
    """A wrapper class for VertexAI embeddings."""
    model_name: str = "text-embedding-004"

    def _create_embedding(self, model_name: str = None, project_id: str = None, region: str = None):
        # Embedding Model:
        return VertexAIEmbeddings(
            model_name=model_name or self.model_name,
            project=project_id or VERTEX_PROJECT_ID,
            location=region or VERTEX_REGION,
            request_parallelism=5,
            max_retries=4,
        )



from typing import Optional, Union, Dict, Any
from collections.abc import Callable
import random
import os
import asyncio
import ssl
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from io import BytesIO
from email.message import Message
from urllib import parse
from urllib.parse import quote, urlencode, urlparse
import urllib3
import aiofiles
# parsing:
from bs4 import BeautifulSoup as bs
from lxml import html, etree
# backoff retries:
import backoff
# aiohttp:
import aiohttp
from aiohttp import BasicAuth
# httpx
import httpx
# config:
from datamodel.typedefs import SafeDict
from datamodel.parsers.json import JSONContent, json_encoder  # pylint: disable=E0611
from navconfig.logging import logging
from proxylists.proxies import (
    FreeProxy,
    Oxylabs
)
from ..conf import (
    HTTPCLIENT_MAX_SEMAPHORE,
    HTTPCLIENT_MAX_WORKERS,
    GOOGLE_SEARCH_API_KEY,
    GOOGLE_SEARCH_ENGINE_ID
)


logging.getLogger("urllib3").setLevel(logging.WARNING)
urllib3.disable_warnings()
logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("httpcore").setLevel(logging.WARNING)
logging.getLogger("aiohttp").setLevel(logging.WARNING)
logging.getLogger("hpack").setLevel(logging.INFO)


ua = [
    # Chrome - Desktop (Windows)
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    # Chrome - Desktop (Mac)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",  # noqa
    # Safari - Desktop (Mac)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.0 Safari/605.1.15",  # noqa
    # Firefox - Desktop (Windows)
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/118.0",
    # Edge - Desktop (Windows)
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36 Edg/118.0.2088.46",  # noqa
    # Chrome - Mobile (Android)
    "Mozilla/5.0 (Linux; Android 13; SM-G991B) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Mobile Safari/537.36",  # noqa
    # Safari - Mobile (iOS)
    "Mozilla/5.0 (iPhone; CPU iPhone OS 16_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Mobile/15E148 Safari/604.1",  # noqa
    # Samsung Internet - Mobile (Android)
    "Mozilla/5.0 (Linux; Android 13; SAMSUNG SM-G991B) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/21.0 Chrome/118.0.0.0 Mobile Safari/537.36",  # noqa
    # Firefox - Mobile (Android)
    "Mozilla/5.0 (Android 13; Mobile; rv:118.0) Gecko/118.0 Firefox/118.0",
    # Opera - Desktop (Windows)
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36 OPR/104.0.0.0"  # noqa
    # Firefox - Desktop (Linux)
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:133.0) Gecko/20100101 Firefox/133.0",
    # Chrome - Desktop (Linux)
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36",
    # Other:
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",  # noqa
]  # noqa

mobile_ua = [
    "Mozilla/5.0 (Linux; Android 4.2.1; en-us; Nexus 5 Build/JOP40D) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.166 Mobile Safari/535.19",  # noqa
    'Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1',  # noqa
    'Mozilla/5.0 (Linux; Android 9; SM-G973F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.119 Mobile Safari/537.36',  # noqa
    'Mozilla/5.0 (Linux; Android 8.0.0; Pixel 2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.93 Mobile Safari/537.36',  # noqa
    'Mozilla/5.0 (Linux; Android 10; HUAWEI VOG-L29) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.138 Mobile Safari/537.36',  # noqa
    'Mozilla/5.0 (iPad; CPU OS 13_2_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0 Mobile/15E148 Safari/604.1',  # noqa
]

impersonates = (
    "chrome_100", "chrome_101", "chrome_104", "chrome_105", "chrome_106", "chrome_107",
    "chrome_108", "chrome_109", "chrome_114", "chrome_116", "chrome_117", "chrome_118",
    "chrome_119", "chrome_120", "chrome_123", "chrome_124", "chrome_126", "chrome_127",
    "chrome_128", "chrome_129", "chrome_130", "chrome_131",
    "safari_ios_16.5", "safari_ios_17.2", "safari_ios_17.4.1", "safari_ios_18.1.1",
    "safari_15.3", "safari_15.5", "safari_15.6.1", "safari_16", "safari_16.5",
    "safari_17.0", "safari_17.2.1", "safari_17.4.1", "safari_17.5",
    "safari_18", "safari_18.2",
    "safari_ipad_18",
    "edge_101", "edge_122", "edge_127", "edge_131",
    "firefox_109", "firefox_117", "firefox_128", "firefox_133",
)  # fmt: skip

impersonates_os = ("android", "ios", "linux", "macos", "windows")


accept_list = {
    "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",  # noqa
    "text/html",
    "application/xhtml+xml",
    "application/xml",
    "image/webp",
    "image/apng",
    "*/*",
    "application/signed-exchange",
    "application/json",
}

valid_methods = ['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD', 'OPTIONS']


def bad_gateway_exception(exc):
    """Check if the exception is a 502 Bad Gateway error."""
    return isinstance(exc, httpx.HTTPStatusError) and exc.response.status_code == 502


class HTTPService:
    """
    Abstraction Class for working with HTTP Clients.

    - aiohttp Client
    - HTTPx
    - Requests
    """
    def __init__(self, *args, **kwargs):
        self.use_proxy: bool = kwargs.pop("use_proxy", False)
        self._free_proxy: bool = kwargs.pop('free_proxy', False)
        self._proxies: list = []
        self.rotate_ua: bool = kwargs.pop("rotate_ua", True)
        self.use_async: bool = bool(kwargs.pop("use_async", True))
        self.google_api_key: str = kwargs.pop('google_api_key', GOOGLE_SEARCH_API_KEY)
        self.google_cse: str = kwargs.pop('google_cse', GOOGLE_SEARCH_ENGINE_ID)
        self.headers: dict = kwargs.get('headers', {})
        self.accept: str = kwargs.get('accept', "application/json")
        self.timeout: int = kwargs.get('timeout', 30)
        self.use_streams: bool = kwargs.get('use_streams', True)
        self.as_binary: bool = kwargs.get('as_binary', False)
        self.no_errors: dict = kwargs.get('no_errors', {})
        self._httpclient: Callable = kwargs.get('httpclient', None)
        self._ua: str = ""
        if self.rotate_ua is True:
            self._ua = random.choice(ua)
        else:
            self._ua: str = ua[0]
        self.headers = {
            "Accept": self.accept,
            "Accept-Encoding": "gzip, deflate",
            "DNT": "1",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": self._ua,
            **self.headers,
        }
        # potentially cookies to be used by request.
        self.cookies = kwargs.get('cookies', {})
        self._encoder = JSONContent()
        # other arguments:
        self.arguments = kwargs
        # Executor:
        self._executor = ThreadPoolExecutor(
            max_workers=int(HTTPCLIENT_MAX_WORKERS)
        )
        self._semaphore = asyncio.Semaphore(
            int(HTTPCLIENT_MAX_SEMAPHORE)
        )
        # Authentication information:
        self.auth: dict = {}
        self.auth_type: str = None
        self.token_type: str = "Bearer"
        self._user, self._pwd = None, None
        self.credentials: dict = kwargs.get('credentials', {})
        if "apikey" in self.credentials:
            self.auth_type = "api_key"
        elif "username" in self.credentials:
            self.auth_type = "basic"
            self._user = self.credentials["username"]
            self._pwd = self.credentials["password"]
        elif "token" in self.credentials:
            self.auth_type = "token"
        elif "key" in self.credentials:
            self.auth_type = "key"
        elif "auth" in self.credentials:
            self.auth_type = "auth"
        # Debugging:
        self._debug: bool = kwargs.pop('debug', False)
        # Parser to be used:
        self._default_parser: str = kwargs.pop('bs4_parser', 'html.parser')
        # Logger:
        self.logger = logging.getLogger(__name__)

    async def get_proxies(self, session_time: float = 1):
        """
        Asynchronously retrieves a list of free proxies.
        TODO: SELECT or rotate the free/paid proxies.
        """
        if self._free_proxy is True:
            return await FreeProxy().get_list()
        else:
            return await Oxylabs(
                session_time=session_time,
                timeout=10
            ).get_list()

    async def refresh_proxies(self):
        """
        Asynchronously refreshes the list of proxies if proxy usage is enabled.
        """
        if self.use_proxy is True:
            self._proxies = await self.get_proxies()

    def build_url(
        self,
        url: str,
        queryparams: Optional[str] = None,
        args: Optional[dict] = None,
        params: Optional[dict] = None
    ) -> str:
        """
        Constructs a full URL with optional query parameters and arguments.

        Args:
            url (str): The base URL.
            queryparams (Optional[str]): The query parameters to append to the URL.
            args (Optional[dict]): Additional arguments to format into the URL.
            params (Optional[dict]): Additional query parameters to append to the URL.

        Returns:
            str: The constructed URL.
        """
        if args:
            url = str(url).format_map(SafeDict(**args))
        if queryparams is not None:
            if "?" in url:
                url += "&" + queryparams
            else:
                url += "?" + queryparams
        if params is not None:
            if "?" in url:
                url += "&" + urlencode(params)
            else:
                url = f"{url}?{urlencode(params)}"
        self.logger.debug(
            f"URL: {url}"
        )
        return url

    def extract_host(self, url):
        """
        Extracts the host from a URL.
        """
        parsed_url = urlparse(url)
        return parsed_url.netloc

    @backoff.on_exception(
        backoff.expo,
        (httpx.HTTPStatusError, httpx.TimeoutException),  # Catch HTTP errors and timeouts
        max_tries=3,
        max_time=120,
        jitter=backoff.full_jitter,
        on_backoff=lambda details: logging.warning(
            f"Retrying HTTP Get: attempt {details['tries']} after {details['wait']:0.2f}s"
        ),
        giveup=lambda e: isinstance(e, httpx.HTTPStatusError) and e.response.status_code not in [429, 500, 502, 503, 504]  # pylint: disable=C0301
    )
    async def _request(
        self,
        url: str,
        method: str = 'get',
        cookies: Optional[httpx.Cookies] = None,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Union[int, float] = 30.0,
        use_proxy: bool = True,
        free_proxy: bool = False,
        use_ssl: bool = True,
        use_json: bool = False,
        follow_redirects: bool = True,
        raise_for_status: bool = True,
        full_response: bool = False,
        connect_timeout: Union[int, float] = 5.0,
        read_timeout: Union[int, float] = 20.0,
        write_timeout: Union[int, float] = 5.0,
        pool_timeout: Union[int, float] = 20.0,
        num_retries: int = 2,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Make an asynchronous HTTPx request, returning the response object.

        Args:
            url (str): The URL to send the request to.
            method (str): The HTTP method to use (default: 'get').
            headers (dict, optional): Dictionary of HTTP headers to include in the request.
            cookies (httpx.Cookies, optional): Cookies to include in the request.
            params (dict, optional): Dictionary of query parameters to include in the URL.
            data (dict, optional): Dictionary of data to send in the request body.
            timeout (float, optional): Total timeout for the request in seconds.
            use_proxy (bool): Whether to use a proxy for the request.
            free_proxy (bool): Whether to use a free proxy.
            use_ssl (bool): Whether to use SSL for the request.
            use_json (bool): Whether to send data as JSON.
            follow_redirects (bool): Whether to follow redirects.
            raise_for_status (bool): Whether to raise an exception for HTTP errors.
            full_response (bool): Whether to return the full response object.
            connect_timeout (float): Timeout for connecting to the server.
            read_timeout (float): Timeout for reading from the server.
            write_timeout (float): Timeout for writing to the server.
            pool_timeout (float): Timeout for connection pool operations.
            num_retries (int): Number of retries to attempt at the transport level.
            **kwargs: Additional arguments to pass to httpx.AsyncClient.

        Returns:
            Tuple[Any, Optional[Dict[str, Any]]]: A tuple containing the result and any error information.

        Raises:
            httpx.TimeoutException: When the request times out.
            httpx.TooManyRedirects: When too many redirects are encountered.
            httpx.HTTPStatusError: When an HTTP error status is encountered.
            httpx.HTTPError: When an HTTP-related error occurs.
            AttributeError: When the HTTP method is invalid.
            RuntimeError: When an unknown error occurs.
        """
        proxy_config = None
        if use_proxy is True:
            self._free_proxy = free_proxy
            proxies = await self.get_proxies()
            if not proxies and use_proxy:
                self.logger.warning(
                    "No proxies available despite use_proxy=True"
                )
            proxy = proxies[0] if proxies else None  # Ensure there's at least one proxy
            proxy_config = {
                "http://": f"http://{proxy}" if proxy else None,
                "https://": f"http://{proxy}" if proxy else None  # Using the same proxy for HTTPS
            }

            # Remove proxies with None values
            proxy_config = {k: v for k, v in proxy_config.items() if v is not None}

        ssl_context = None
        if use_ssl:
            # Define custom SSL context
            ssl_context = ssl.create_default_context()
            # Disable older protocols if needed
            ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
            # Ensure at least TLS 1.2 is used
            ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
            # Make this configurable rather than hardcoded to CERT_NONE
            if kwargs.get('verify_ssl', True):
                ssl_context.check_hostname = True
                ssl_context.verify_mode = ssl.CERT_REQUIRED
            else:
                ssl_context.check_hostname = False
                ssl_context.verify_mode = ssl.CERT_NONE

        # Use AsyncHTTPTransport to pass in SSL context explicitly
        transport_options = {
            'retries': num_retries,
            'verify': ssl_context
        }
        if 'transport_options' in kwargs:
            transport_options.update(kwargs.pop('transport_options'))
        transport = httpx.AsyncHTTPTransport(
            **transport_options
        )
        timeout = httpx.Timeout(
            timeout=timeout,
            connect=connect_timeout,
            read=read_timeout,
            write=write_timeout,
            pool=pool_timeout
        )
        method = method.upper()
        if method not in valid_methods:
            raise ValueError(
                f"Invalid HTTP method: {method}. Must be one of {valid_methods}"
            )
        async with httpx.AsyncClient(
            cookies=cookies,
            proxies=proxy_config or None,
            transport=transport,
            headers=headers,
            timeout=timeout,
            http2=kwargs.get('use_http2', True),
            follow_redirects=follow_redirects,
            **kwargs
        ) as client:
            try:
                args = {
                    "method": method.upper(),
                    "url": url,
                    "follow_redirects": follow_redirects
                }
                if data:
                    if use_json:
                        args["json"] = data
                    else:
                        args["data"] = data
                if params:
                    args["params"] = params
                if self._httpclient:
                    # keep session alive.
                    response = await client.request(
                        **args
                    )
                else:
                    response = await client.request(**args)
                if raise_for_status:
                    response.raise_for_status()
                if full_response:
                    if self.logger.isEnabledFor(logging.DEBUG):
                        self.logger.debug(
                            f"Response from {url}: status={response.status_code}, headers={response.headers}"
                        )
                    return response, None
                result, error = await self.process_response(
                    response,
                    url,
                    download=kwargs.get('download', False),
                    filename=kwargs.get('filename', None)
                )
                return result, error
            except httpx.TimeoutException:
                self.logger.error("Request timed out.")
                raise
            except httpx.TooManyRedirects:
                self.logger.error("Too many redirects.")
                raise
            except httpx.HTTPStatusError as ex:
                self.logger.error(
                    f"HTTP status error occurred: {ex.response.status_code} - {ex}"
                )
                raise
            except httpx.HTTPError as ex:
                self.logger.error(f"HTTP error occurred: {ex}")
                raise
            except AttributeError as e:
                self.logger.error(f"HTTPx Client doesn't have attribute {method}: {e}")
                raise
            except Exception as exc:
                self.logger.error(f'Unknown Error > {exc}')
                raise RuntimeError(
                    f"An error occurred: {exc}"
                ) from exc

    @backoff.on_exception(
        backoff.expo,                     # Use exponential backoff
        (aiohttp.ClientError,             # Retry on network-related errors
        aiohttp.ServerTimeoutError,      # Retry on timeouts
        aiohttp.ClientResponseError),    # Retry on certain HTTP errors
        max_tries=3,                      # Maximum number of retries
        max_time=60,                      # Maximum total time to try (in seconds)
        jitter=backoff.full_jitter,       # Use full jitter to randomize retry intervals
        giveup=lambda e: isinstance(e, aiohttp.ClientResponseError) and e.status not in [429, 500, 502, 503, 504]  # pylint: disable=C0301
    )
    async def async_request(
        self,
        url,
        method: str = 'GET',
        data: dict = None,
        use_ssl: bool = False,
        use_json: bool = False,
        use_proxy: bool = False,
        accept: Optional[str] = None,
        download: bool = False,
        full_response: bool = False
    ):
        """
        Asynchronously sends an HTTP request using aiohttp.

        :param url: The URL to send the request to.
        :param method: The HTTP method to use (e.g., 'GET', 'POST').
        :param data: The data to send in the request body.
        :param use_json: Whether to send the data as JSON.
        :param use_proxy: force proxy usage.
        :param accept: The accept header to use.
        :param download: Whether to download the response as a file.
        :param full_response: Whether to return the full response object or result processed.
        :return: A tuple containing the result and any error information.
        """
        result = []
        error = {}
        auth = None
        proxy = None
        ssl_context = None

        if use_proxy is True:
            self._proxies = await self.get_proxies()
        if self._proxies:
            proxy = random.choice(self._proxies)
        if self.credentials:
            if "apikey" in self.auth:
                self.headers[
                    "Authorization"
                ] = f"{self.token_type} {self.auth['apikey']}"
            elif self.auth_type == "api_key":
                self.headers = {**self.headers, **self.credentials}
            elif self.auth_type == "key":
                url = self.build_url(
                    url,
                    args=self.arguments,
                    queryparams=urlencode(self.credentials)
                )
            elif self.auth_type in ["basic", "auth", "user"]:
                auth = BasicAuth(
                    self.credentials["username"],
                    self.credentials["password"]
                )
        elif "apikey" in self.auth:
            self.headers["Authorization"] = f"{self.token_type} {self.auth['apikey']}"
        elif self.auth:
            token_type, token = list(self.auth.items())[0]
            self.headers["Authorization"] = f"{token_type} {token}"
        elif self._user and self.auth_type == "basic":
            auth = BasicAuth(self._user, self._pwd)
        self.logger.debug(
            f"HTTP: Connecting to {url} using {method}",
            level="DEBUG"
        )
        if auth is not None:
            args = {"auth": auth}
        else:
            args = {}
        if use_ssl:
            ssl_context = ssl.create_default_context()
            # Disable older protocols if needed
            ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
            # Ensure at least TLS 1.2 is used
            ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE
            args['ssl'] = True
            args['ssl_context'] = ssl_context
        if accept is not None:
            self.headers["Accept"] = accept
        else:
            self.headers["Accept"] = self.accept
        if download is True:
            self.headers["Accept"] = "application/octet-stream"
            self.headers["Content-Type"] = "application/octet-stream"
            if hasattr(self, "use_streams"):
                self.headers["Transfer-Encoding"] = "chunked"
                args["stream"] = True
        timeout = aiohttp.ClientTimeout(total=self.timeout)
        async with aiohttp.ClientSession(
            headers=self.headers,
            timeout=timeout,
            auth=auth,
            json_serialize=json_encoder,
        ) as session:
            try:
                if use_json is True:
                    async with session.request(
                        method.upper(), url, json=data, proxy=proxy, **args
                    ) as response:
                        if full_response is True:
                            return full_response, None
                        result, error = await self.process_response(response, url, download=download)
                else:
                    async with session.request(
                        method.upper(), url, data=data, proxy=proxy, **args
                    ) as response:
                        if full_response is True:
                            return full_response, None
                        # Process the response
                        result, error = await self.process_response(response, url, download=download)
            except aiohttp.ClientError as e:
                error = str(e)
        return (result, error)

    async def evaluate_error(
        self, response: Union[str, list], message: Union[str, list, dict]
    ) -> tuple:
        """evaluate_response.

        Check Response status and available payloads.
        Args:
            response (_type_): _description_
            url (str): _description_

        Returns:
            tuple: _description_
        """
        if isinstance(response, list):
            # a list of potential errors:
            for msg in response:
                if message in msg:
                    return True
        if isinstance(response, dict) and "errors" in response:
            errors = response["errors"]
            if isinstance(errors, list):
                for error in errors:
                    try:
                        if message in error:
                            return True
                    except TypeError:
                        if message == error:
                            return True
            else:
                if message == errors:
                    return True
        else:
            if message in response:
                return True
        return False

    def response_status(self, response):
        if hasattr(response, 'status_code'):
            return response.status_code
        return response.status

    async def response_json(self, response):
        if asyncio.iscoroutinefunction(response.json):
            return await response.json()
        return response.json()

    async def response_read(self, response):
        if hasattr(response, 'aread'):
            return await response.aread()
        return await response.read()

    async def response_text(self, response):
        if asyncio.iscoroutinefunction(response.text):
            return await response.text()
        return response.text

    async def response_reason(self, response):
        # Attempt to retrieve `reason`, `reason_phrase`, or fallback to an empty string
        reason = getattr(response, 'reason', getattr(response, 'reason_phrase', b''))
        return f"{reason!s}"

    async def process_response(
        self,
        response, url: str,
        download: bool = False,
        filename: Optional[str] = None
    ) -> tuple:
        """
        Processes the response from an HTTP request.

        :param response: The response object from aiohttp.
        :param url: The URL that was requested.
        :param download: Whether to download the response as a file.
        :param filename: The filename to use for downloading the response.
        :return: A tuple containing the processed result and any error information.
        """
        error = None
        result = None
        # Process the response
        status = self.response_status(response)

        if status >= 400:
            # Evaluate response body and headers.
            print(" == ERROR Headers == ")
            print(f"{response.headers}")
            content_type = response.headers.get("Content-Type", "").lower()
            if "application/json" in content_type:
                message = await self.response_json(response)
            elif "text/" in content_type:
                message = await self.response_text(response)
            elif "X-Error" in response.headers:
                message = response.headers["X-Error"]
            else:
                # Fallback to a unified read method for the raw body content
                message = await self.response_read(response)

            # Combine response headers and body for enriched logging
            error_context = {
                "status": status,
                "reason": await self.response_reason(response),
                "headers": response.headers,
                "body": message
            }

            # Log the detailed error context
            self.logger.error(
                f"Error: {error_context}"
            )

            # Additional error handling or custom evaluation based on status
            if self.no_errors:
                for key, msg in self.no_errors.items():
                    if int(key) == status and await self.evaluate_error(message, msg):
                        return response, status

            # Raise an exception if error handling does not continue
            raise ConnectionError(
                f"HTTP Error {status}: {message!s}"
            )
        else:
            if download is True:
                if not filename:
                    filename = os.path.basename(url)
                # Get the filename from the response headers, if available
                content_disposition = response.headers.get("content-disposition")
                if content_disposition:
                    msg = Message()
                    msg["Content-Disposition"] = response.headers.get("content-disposition")
                    filename = msg.get_param("filename", header="Content-Disposition")
                    utf8_filename = msg.get_param("filename*", header="Content-Disposition")
                    if utf8_filename:
                        _, utf8_filename = utf8_filename.split("''", 1)
                        filename = parse.unquote(utf8_filename)
                if "{filename}" in str(filename):
                    filename = str(filename).format_map(
                        SafeDict(filename=filename)
                    )
                if "{" in str(filename):
                    filename = str(filename).format_map(
                        SafeDict(**self.arguments)
                    )
                if isinstance(filename, str):
                    filename = Path(filename)
                # Saving File in Directory:
                total_length = response.headers.get("Content-Length")
                self.logger.info(
                    f"HTTPClient: Saving File {filename}, size: {total_length}"
                )
                pathname = filename.parent.absolute()
                if not pathname.exists():
                    # Create a new directory
                    pathname.mkdir(parents=True, exist_ok=True)
                transfer = response.headers.get("transfer-encoding", None)
                if transfer is None:
                    chunk_size = int(total_length)
                else:
                    chunk_size = 8192
                # Asynchronous file writing
                if filename.exists() and filename.is_file():
                    self.logger.warning(
                        f"HTTPClient: File Already exists: {filename}"
                    )
                    # Filename already exists
                    result = filename
                    return result, error
                if self.use_streams is True:
                    async with aiofiles.open(filename, "wb") as file:
                        async for chunk in response.content.iter_chunked(chunk_size):
                            await file.write(chunk)
                else:
                    with open(filename, "wb") as fp:
                        try:
                            fp.write(await self.response_read(response))
                        except Exception:
                            pass
                self.logger.debug(
                    f"Filename Saved Successfully: {filename}"
                )
                result = filename
            else:
                if self.accept == 'application/octet-stream':
                    data = await self.response_read(response)
                    buffer = BytesIO(data)
                    buffer.seek(0)
                    result = buffer
                elif self.accept in ('text/html'):
                    result = await self.response_read(response)
                    try:
                        # html parser for lxml
                        self._parser = html.fromstring(result)
                        # BeautifulSoup parser
                        self._bs = bs(response.text, self._default_parser)
                        result = self._bs
                    except Exception as e:
                        error = e
                elif self.accept in ('application/xhtml+xml', 'application/xml'):
                    result = await self.response_read(response)
                    try:
                        self._parser = etree.fromstring(result)  # pylint: disable=I1101
                    except etree.XMLSyntaxError:  # pylint: disable=I1101
                        self._parser = html.fromstring(result)
                    except Exception as e:
                        error = e
                elif self.accept == "application/json":
                    try:
                        result = await self.response_json(response)
                    except Exception as e:
                        logging.error(e)
                        # is not an json, try first with beautiful soup:
                        try:
                            self._bs = bs(
                                await self.response_text(response),
                                self._default_parser
                            )
                            result = self._bs
                        except Exception:
                            error = e
                elif self.as_binary is True:
                    result = await self.response_read(
                        response
                    )
                else:
                    result = await self.response_text(
                        response
                    )
        return result, error



import os
from navconfig import config, BASE_DIR
from google.cloud import aiplatform
import vertexai
from vertexai.preview.vision_models import ImageGenerationModel
from langchain_google_vertexai import (
    ChatVertexAI,
    VertexAI,
    VertexAIModelGarden,
)
from .abstract import AbstractLLM

class VertexLLM(AbstractLLM):
    """VertexLLM.

    Interact with VertexAI Language Model.

    Returns:
        _type_: VertexAI LLM.
    """
    model: str = "gemini-2.0-pro"
    max_tokens: int = 1024
    supported_models: list = [
        "gemini-1.0-pro",
        "gemini-1.5-pro-001",
        "gemini-1.5-pro-exp-0801",
        "gemini-1.5-flash-preview-0514",
        "gemini-1.5-flash-001",
        "chat-bison@001",
        "chat-bison@002",
        "claude-3-opus@20240229",
        'claude-3-5-sonnet@20240620'
    ]

    def __init__(self, *args, use_chat: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        use_garden: bool = kwargs.get("use_garden", False)
        project_id = config.get("VERTEX_PROJECT_ID")
        region = config.get("VERTEX_REGION")
        # vertexai.init(project=project_id, location="us-central1")
        config_file = config.get('GOOGLE_CREDENTIALS_FILE', 'env/google/vertexai.json')
        config_dir = BASE_DIR.joinpath(config_file)
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = str(config_dir)
        self.args = {
            "project": project_id,
            "location": region,
            "max_output_tokens": self.max_tokens,
            "temperature": self.temperature,
            "max_retries": 4,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "verbose": True,
        }
        if use_garden is True:
            base_llm = VertexAIModelGarden
            self._llm = VertexAIModelGarden(
                endpoint_id=self.model,
                **self.args
            )
        else:
            if use_chat is True:
                base_llm = ChatVertexAI
            else:
                base_llm = VertexAI
            self._llm = base_llm(
                model_name=self.model,
                system_prompt="Always respond in the same language as the user's question. If the user's language is not English, translate your response into their language.",
                **self.args
            )
        # LLM
        self._version_ = aiplatform.__version__



from navconfig import BASE_DIR
from crewai_tools import BaseTool
import markdown2
from weasyprint import HTML


class MarkdownToPDFTool(BaseTool):
    """Markdown to PDF Tool."""
    name: str = "MarkdownToPDFTool"
    description: str = "Converts markdown documents to PDF format."

    def _run(self, markdown_content: str) -> str:
        # Convert Markdown to HTML
        html_content = markdown2.markdown(markdown_content)
        # Convert HTML to PDF (assuming it's a simple conversion without advanced styling)
        pdf = HTML(string=html_content).write_pdf()

        # Save PDF to a temporary file and return file path
        pdf_file_path = BASE_DIR.joinpath('docs', 'report.pdf')
        with open(str(pdf_file_path), 'wb') as f:
            f.write(pdf)
        return pdf_file_path

    def __call__(self, markdown_content: str) -> str:
        return self._run(markdown_content)



from typing import Optional
import requests
from googleapiclient.discovery import build
from pydantic import PrivateAttr
# from crewai_tools import BaseTool
from langchain.tools import BaseTool
from langchain.tools import Tool
from navconfig import config
from ...conf import GOOGLE_API_KEY


class GoogleSearchTool(BaseTool):
    """Web Search tool using Google API."""
    name: str = "Google Web Search"
    description: str = "Search the web using Google Search API, useful when you need to answer questions about current events."
    source: str = 'news'
    max_results: int = 5
    region: str = ''
    # Fields populated during init (not required for validation)
    cse_id: Optional[str] = None
    search_key: Optional[str] = None
    kwargs: Optional[dict] = None

    def __init__(self, source: str = "news", results: int = 5, **kwargs):
        super().__init__(**kwargs)
        self.source = source
        self.max_results = results
        self.cse_id = config.get('GOOGLE_SEARCH_ENGINE_ID')
        self.search_key = config.get('GOOGLE_SEARCH_API_KEY')
        self.kwargs = kwargs

    def _run(self, query: str) -> list:
        """Run the Google Search Tool."""
        service = build("customsearch", "v1", developerKey=self.search_key)
        res = service.cse().list(
            q=query,
            cx=self.cse_id,
            num=self.max_results,
            **self.kwargs
        ).execute()
        results = []
        for item in res['items']:
            results.append(
                {
                    'snippet': item['snippet'],
                    'title': item['title'],
                    'link': item['link'],
                    'description': item['snippet']
                }
            )
        return results


class GoogleSiteSearchTool(BaseTool):
    """Web Search under a site using Google API."""
    name: str = "Google Web Search"
    description: str = "Search the web using Google Search API"
    source: str = 'news'
    max_results: int = 5
    region: str = ''

    def __init__(self, site: str = "news", results: int = 5, **kwargs):
        super().__init__(**kwargs)
        self.source = site
        self.max_results = results
        self._cse_id = config.get('GOOGLE_SEARCH_ENGINE_ID')
        self._search_key = config.get('GOOGLE_SEARCH_API_KEY')
        self._kwargs = kwargs

    def _run(self, query: str) -> dict:
        """Run the Google Search Tool."""
        service = build("customsearch", "v1", developerKey=self._search_key)
        qs = f'{query} site:{self.source}'
        res = service.cse().list(
            q=qs,
            cx=self._cse_id,
            num=self.max_results,
            **self._kwargs
        ).execute()
        results = []
        for item in res['items']:
            results.append(
                {
                    'snippet': item['snippet'],
                    'title': item['title'],
                    'link': item['link'],
                    'description': item['snippet']
                }
            )
        return results


class GoogleLocationFinder(BaseTool):
    """ LocationFinder class for finding locations."""
    name: str = "google_maps_location_finder"
    description: str = (
        "Search for location information, use this tool to find latitude, longitude and other geographical information from locations."
        " Provide the complete address to this tool to receive location information"
    )
    google_key: str = None
    base_url: str = "https://maps.googleapis.com/maps/api/geocode/json"
    kwargs: Optional[dict] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.google_key = kwargs.get('api_key', GOOGLE_API_KEY)
        self.kwargs = kwargs

    def extract_location(self, data):
        city = state = state_code = zipcode = None
        try:
            for component in data['address_components']:
                if 'locality' in component['types']:
                    city = component['long_name']
                elif 'administrative_area_level_1' in component['types']:
                    state_code = component['short_name']
                    state = component['long_name']
                elif 'postal_code' in component['types']:
                    zipcode = component['long_name']
        except Exception:
            pass
        return city, state, state_code, zipcode

    def _run(self, query: str) -> dict:
        """Find Location."""
        params = {
            "address": query,
            "key": self.google_key
        }
        response = requests.get(
            self.base_url,
            params=params
        )
        if response.status_code == 200:
            result = response.json()
            if result['status'] == 'OK':
                location = result['results'][0]
                city, state, state_code, zipcode = self.extract_location(
                    location
                )
                return  {
                    "latitude": location['geometry']['location']['lat'],
                    "longitude": location['geometry']['location']['lng'],
                    "address": location['formatted_address'],
                    "place_id": location['place_id'],
                    "zipcode": zipcode,
                    "city": city,
                    "state": state,
                    "state_code": state_code
                }
            return None
        else:
            return None

class GoogleRouteSearch(BaseTool):
    """Web Search under a site using Google API."""
    name: str = "google_maps_route_search"
    description: str = "Search for a Route to a location using Google Maps, using this tool if answers questions about how to reach a location."
    google_key: str = None
    base_url: str = 'https://maps.googleapis.com/maps/api/directions/json'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._key_ = kwargs.get('api_key', GOOGLE_API_KEY)
        self._kwargs = kwargs

    def _run(self, query: str) -> dict:
        departure_time = 'now'



from pathlib import Path
from navconfig import config, BASE_DIR
from navconfig.logging import logging
from navigator.conf import default_dsn, CACHE_HOST, CACHE_PORT


# disable debug on some libraries:
logging.getLogger(name='httpcore').setLevel(logging.INFO)
logging.getLogger(name='httpx').setLevel(logging.INFO)
logging.getLogger(name='groq').setLevel(logging.INFO)
logging.getLogger(name='tensorflow').setLevel(logging.INFO)
logging.getLogger(name='selenium.webdriver').setLevel(logging.WARNING)
logging.getLogger(name='selenium').setLevel(logging.INFO)
logging.getLogger(name='matplotlib').setLevel(logging.WARNING)
logging.getLogger(name='PIL').setLevel(logging.INFO)


# Static directory
STATIC_DIR = config.get('STATIC_DIR', fallback=BASE_DIR.joinpath('static'))
if isinstance(STATIC_DIR, str):
    STATIC_DIR = Path(STATIC_DIR)

# LLM Model
DEFAULT_LLM_MODEL_NAME = config.get('LLM_MODEL_NAME', fallback='gemini-pro')


## MILVUS DB ##:
MILVUS_HOST = config.get('MILVUS_HOST', fallback='localhost')
MILVUS_PROTOCOL = config.get('MILVUS_PROTOCOL', fallback='http')
MILVUS_PORT = config.get('MILVUS_PORT', fallback=19530)
MILVUS_URL = config.get('MILVUS_URL')
MILVUS_TOKEN = config.get('MILVUS_TOKEN')
MILVUS_USER = config.get('MILVUS_USER')
MILVUS_PASSWORD = config.get('MILVUS_PASSWORD')
MILVUS_SECURE = config.getboolean('MILVUS_SECURE', fallback=False)
MILVUS_SERVER_NAME = config.get(
    'MILVUS_SERVER_NAME'
)
MILVUS_CA_CERT = config.get('MILVUS_CA_CERT', fallback=None)
MILVUS_SERVER_CERT = config.get('MILVUS_SERVER_CERT', fallback=None)
MILVUS_SERVER_KEY = config.get('MILVUS_SERVER_KEY', fallback=None)
MILVUS_USE_TLSv2 = config.getboolean('MILVUS_USE_TLSv2', fallback=False)

# ScyllaDB Database:
SCYLLADB_DRIVER = config.get('SCYLLADB_DRIVER', fallback='scylladb')
SCYLLADB_HOST = config.get('SCYLLADB_HOST', fallback='localhost')
SCYLLADB_PORT = int(config.get('SCYLLADB_PORT', fallback=9042))
SCYLLADB_USERNAME = config.get('SCYLLADB_USERNAME', fallback='navigator')
SCYLLADB_PASSWORD = config.get('SCYLLADB_PASSWORD', fallback='navigator')
SCYLLADB_KEYSPACE = config.get('SCYLLADB_KEYSPACE', fallback='navigator')


# BigQuery Configuration:
BIGQUERY_CREDENTIALS = config.get('BIGQUERY_CREDENTIALS')
BIGQUERY_PROJECT_ID = config.get('BIGQUERY_PROJECT_ID', fallback='navigator')
BIGQUERY_DATASET = config.get('BIGQUERY_DATASET', fallback='navigator')

# Redis History Configuration:
REDIS_HISTORY_DB = config.get('REDIS_HISTORY_DB', fallback=3)
REDIS_HISTORY_URL = f"redis://{CACHE_HOST}:{CACHE_PORT}/{REDIS_HISTORY_DB}"

def resolve_cert(crt):
    cert = Path(crt)
    if not cert.is_absolute():
        cert = BASE_DIR.joinpath(cert)
    else:
        cert.resolve()
    return cert

if MILVUS_SERVER_CERT:
    MILVUS_SERVER_CERT = str(resolve_cert(MILVUS_SERVER_CERT))
if MILVUS_CA_CERT:
    MILVUS_CA_CERT = str(resolve_cert(MILVUS_CA_CERT))
if MILVUS_SERVER_KEY:
    MILVUS_SERVER_KEY = str(resolve_cert(MILVUS_SERVER_KEY))

# QDRANT:
QDRANT_PROTOCOL = config.get('QDRANT_PROTOCOL', fallback='http')
QDRANT_HOST = config.get('QDRANT_HOST', fallback='localhost')
QDRANT_PORT = config.get('QDRANT_PORT', fallback=6333)
QDRANT_USE_HTTPS = config.getboolean('QDRANT_USE_HTTPS', fallback=False)
QDRANT_URL = config.get('QDRANT_URL')
# QDRANT Connection Type: server or cloud
QDRANT_CONN_TYPE = config.get('QDRANT_CONN_TYPE', fallback='server')

# ChromaDB:
CHROMADB_HOST = config.get('CHROMADB_HOST', fallback='localhost')
CHROMADB_PORT = config.get('CHROMADB_PORT', fallback=8000)

# Embedding Device:
EMBEDDING_DEVICE = config.get('EMBEDDING_DEVICE', fallback='cpu')
EMBEDDING_DEFAULT_MODEL = config.get(
    'EMBEDDING_DEFAULT_MODEL',
    fallback='thenlper/gte-base'
)
MAX_VRAM_AVAILABLE = config.get('MAX_VRAM_AVAILABLE', fallback=20000)
RAM_AVAILABLE = config.get('RAM_AVAILABLE', fallback=819200)
CUDA_DEFAULT_DEVICE = config.get('CUDA_DEFAULT_DEVICE', fallback=0)
MAX_BATCH_SIZE = config.get('MAX_BATCH_SIZE', fallback=768)

# Enable Teams Bot:
ENABLE_AZURE_BOT = config.getboolean('ENABLE_AZURE_BOT', fallback=True)

## Google API:
GOOGLE_API_KEY = config.get('GOOGLE_API_KEY')
### Google Service Credentials:
GA_SERVICE_ACCOUNT_NAME = config.get('GA_SERVICE_ACCOUNT_NAME', fallback="google.json")
GA_SERVICE_PATH = config.get('GA_SERVICE_PATH', fallback="env/google/")

# Google SerpAPI:
SERPAPI_API_KEY = config.get('SERPAPI_API_KEY')

# Groq API Key:
GROQ_API_KEY = config.get('GROQ_API_KEY')

# Ethical Principle:
ETHICAL_PRINCIPLE = config.get(
    'ETHICAL_PRINCIPLE',
    fallback='The model should only talk about ethical and legal things.'
)

# Embedding Configuration:

# VERTEX
VERTEX_PROJECT_ID = config.get('VERTEX_PROJECT_ID')
VERTEX_REGION = config.get('VERTEX_REGION')

# OpenAI:
OPENAI_API_KEY = config.get('OPENAI_API_KEY')
OPENAI_ORGANIZATION = config.get('OPENAI_ORGANIZATION')

## HTTPClioent
HTTPCLIENT_MAX_SEMAPHORE = config.getint("HTTPCLIENT_MAX_SEMAPHORE", fallback=5)
HTTPCLIENT_MAX_WORKERS = config.getint("HTTPCLIENT_MAX_WORKERS", fallback=1)

## Google API:
GOOGLE_API_KEY = config.get('GOOGLE_API_KEY')
GOOGLE_SEARCH_API_KEY = config.get('GOOGLE_SEARCH_API_KEY')
GOOGLE_SEARCH_ENGINE_ID = config.get('GOOGLE_SEARCH_ENGINE_ID')
GOOGLE_PLACES_API_KEY = config.get('GOOGLE_PLACES_API_KEY')
GOOGLE_CREDENTIALS_FILE = Path(
    config.get(
        'GOOGLE_CREDENTIALS_FILE',
        fallback=BASE_DIR.joinpath('env', 'google', 'key.json')
    )
)



from langchain_community.embeddings.fastembed import (
    FastEmbedEmbeddings
)
from .abstract import AbstractEmbed


class FastembedEmbed(AbstractEmbed):
    """A wrapper class for FastEmbed embeddings."""
    model_name: str = "BAAI/bge-large-en-v1.5"

    def _create_embedding(self, model_name: str = None, **kwargs):
        # Embedding Model:
        return FastEmbedEmbeddings(
            model_name=model_name,
            max_length=1024,
            threads=4
        )



"""
Collection of useful prompts for Chatbots.
"""
from .agents import AGENT_PROMPT


BASIC_SYSTEM_PROMPT = """
You are {name}, a helpful and professional AI assistant.

Your primary function is to {goal}
Use the information from the provided knowledge base and provided context of documents to answer users' questions accurately and concisely.
Focus on answering the question directly but in detail. Do not include an introduction or greeting in your response.

I am here to help with {role}.

**Backstory:**
{backstory}.

Here is a brief summary of relevant information:
Context: {context}
End of Context.

**{rationale}**

Given this information, please provide answers to the following question adding detailed and useful insights.
"""

BASIC_HUMAN_PROMPT = """
**Chat History:**
{chat_history}

**Human Question:**
{question}
"""

DEFAULT_BACKHISTORY = "I am an AI assistant designed to help you find information.\n"



from typing import Optional
import gc
import torch
from .abstract import AbstractEmbed
from ...conf import CUDA_DEFAULT_DEVICE, EMBEDDING_DEVICE


class BaseEmbed(AbstractEmbed):
    """A wrapper class for Base embeddings.

    Use this class to Embedding Models that requires Torch/Transformers.
    """
    model_kwargs = {
        'device': EMBEDDING_DEVICE,
        'trust_remote_code':True
    }
    def _get_device(self, device_type: str = None, cuda_number: Optional[int] = None):
        """Get Default device for Torch and transformers.

        """
        # torch.backends.cudnn.deterministic = True
        if device_type is not None:
            return torch.device(device_type)
        if torch.cuda.is_available():
            # Use CUDA GPU if available
            if cuda_number is None:
                cuda_number = CUDA_DEFAULT_DEVICE
            return torch.device(f'cuda:{cuda_number}')
        if torch.backends.mps.is_available():
            # Use CUDA Multi-Processing Service if available
            return torch.device("mps")
        if EMBEDDING_DEVICE == 'cuda':
            if cuda_number is None:
                cuda_number = CUDA_DEFAULT_DEVICE
            return torch.device(f'cuda:{cuda_number}')
        else:
            return torch.device(EMBEDDING_DEVICE)

    def free(self):
        """
        Free the resources.
        """
        try:
            gc.collect()               # Run Python garbage collector to free unreferenced objects
            torch.cuda.empty_cache()   # Release cached memory blocks back to the GPU
        except Exception as e:
            print(f"Error: {e}")



from pathlib import Path
from typing import List, Any, Union
import os
from datetime import datetime, timezone
from langchain_core.prompts import (
    PromptTemplate,
    ChatPromptTemplate
)
from langchain_core.retrievers import BaseRetriever
from langchain import hub
from langchain.callbacks.base import BaseCallbackHandler
from langchain.agents import (
    create_react_agent,
    create_openai_functions_agent,
    create_openai_tools_agent,
    create_tool_calling_agent
)
from langchain.agents.agent import (
    AgentExecutor,
    RunnableAgent,
    RunnableMultiActionAgent,
)
from langchain.agents.agent_toolkits import create_retriever_tool
from langchain.prompts import (
    ChatPromptTemplate,
    MessagesPlaceholder,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
)
from langchain_community.utilities import SQLDatabase
from langchain_community.agent_toolkits.sql.toolkit import SQLDatabaseToolkit
from langchain_community.agent_toolkits.sql.base import create_sql_agent
from langchain_community.agent_toolkits.json.base import create_json_agent
from langchain_community.tools.json.tool import JsonSpec
from langchain_community.agent_toolkits.json.toolkit import JsonToolkit
from langchain_community.utilities import TextRequestsWrapper
# for exponential backoff
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)  # for exponential backoff
from datamodel.typedefs import SafeDict
from datamodel.parsers.json import json_decoder  # noqa  pylint: disable=E0611
from navconfig.logging import logging
from .abstract import AbstractBot
from .prompts import AGENT_PROMPT
from ..models import AgentResponse
from ..tools import AbstractTool, SearchTool, MathTool, DuckDuckGoSearchTool


os.environ["GRPC_ENABLE_FORK_SUPPORT"] = "0"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"  # Hide TensorFlow logs if present
logging.getLogger("grpc").setLevel(logging.CRITICAL)

class BasicAgent(AbstractBot):
    """Represents an Agent in Navigator.

        Agents are chatbots that can access to Tools and execute commands.
        Each Agent has a name, a role, a goal, a backstory,
        and an optional language model (llm).

        These agents are designed to interact with structured and unstructured data sources.
    """
    def __init__(
        self,
        name: str = 'Agent',
        agent_type: str = 'zero_shot',
        llm: str = 'vertexai',
        tools: List[AbstractTool] = None,
        system_prompt: str = None,
        human_prompt: str = None,
        prompt_template: str = None,
        **kwargs
    ):
        super().__init__(
            name=name,
            llm=llm,
            system_prompt=system_prompt,
            human_prompt=human_prompt,
            **kwargs
        )
        self.agent = None
        self.agent_type = agent_type
        self._agent = None # Agent Executor
        self.prompt_template = prompt_template or AGENT_PROMPT
        self.tools = tools or self.default_tools(tools)
        if system_prompt:
            self.prompt_template = self.prompt_template.format_map(
                SafeDict(
                    system_prompt_base=system_prompt
                )
            )
        else:
            self.prompt_template = self.prompt_template.format_map(
                SafeDict(
                    system_prompt_base="""
                    Whether you need help with a specific question or just want to have a conversation about a particular topic, Assistant is here to assist.
                    """
                )
            )
        self.prompt = self.define_prompt(self.prompt_template)
        ##  Logging:
        self.logger = logging.getLogger(
            f'{self.name}.Agent'
        )

    def default_tools(self, tools: list = None) -> List[AbstractTool]:
        ctools = [
            DuckDuckGoSearchTool(),
            SearchTool(),
            MathTool()
        ]
        if tools:
            ctools.extend(tools)
        return ctools

    def define_prompt(self, prompt, **kwargs):
        now = datetime.now(timezone.utc).strftime("%Y-%m-%d")
        list_of_tools = ""
        for tool in self.tools:
            name = tool.name
            description = tool.description  # noqa  pylint: disable=E1101
            list_of_tools += f'- {name}: {description}\n'
        list_of_tools += "\n"
        final_prompt = prompt.format_map(
            SafeDict(
                today_date=now,
                list_of_tools=list_of_tools
            )
        )
        # Define a structured system message
        system_message = f"""
        Today is {now}. If an event is expected to have occurred before this date,
        assume that results exist and verify using a web search tool.

        If you call a tool and receive a valid answer, finalize your response immediately.
        Do NOT repeat the same tool call multiple times for the same question.
        """
        chat_prompt = ChatPromptTemplate.from_messages([
            SystemMessagePromptTemplate.from_template(system_message),
            ChatPromptTemplate.from_template(final_prompt)
        ])
        return chat_prompt.partial(
            tools=self.tools,
            tool_names=", ".join([tool.name for tool in self.tools]),
            name=self.name,
            **kwargs
        )

    def get_retriever_tool(
        self,
        retriever: BaseRetriever,
        name: str = 'vector_retriever',
        description: str = 'Search for information about a topic in a Vector Retriever.',
    ):
        return create_retriever_tool(
            name=name,
            description=description,
            retriever=retriever,
        )

    def runnable_json_agent(self, json_file: Union[str, Path], **kwargs):
        """
        Creates a JSON Agent using `create_json_agent`.

        This agent is designed to work with structured JSON input and output.

        Returns:
            RunnableMultiActionAgent: A JSON-based agent.

        ✅ Use Case: Best when dealing with structured JSON data and needing a predictable schema.
        """
        data = None
        if isinstance(json_file, str):
            data = json_file
        elif isinstance(json_file, Path):
            data = json_file.read_text()
        data = json_decoder(data)
        json_spec = JsonSpec(dict_= data, max_value_length=4000)
        json_toolkit = JsonToolkit(spec=json_spec)
        agent = create_json_agent(
            llm=self._llm,
            toolkit=json_toolkit,
            verbose=True,
            prompt=self.prompt,
        )
        return self.prompt | self._llm | agent

    def runnable_agent(self, **kwargs):
        """
        Creates a ZeroShot ReAct Agent.

        This agent uses reasoning and tool execution iteratively to generate responses.

        Returns:
            RunnableMultiActionAgent: A ReAct-based agent.

        ✅ Use Case: Best for decision-making and reasoning tasks where the agent must break problems down into multiple steps.

        """
        return RunnableMultiActionAgent(
            runnable = create_react_agent(
                self._llm,
                self.tools,
                prompt=self.prompt,
            ),  # type: ignore
            input_keys_arg=["input"],
            return_keys_arg=["output"],
            **kwargs
        )

    def function_calling_agent(self, **kwargs):
        """
        Creates a Function Calling Agent.

        This agent uses reasoning and tool execution iteratively to generate responses.

        Returns:
            RunnableMultiActionAgent: A ReAct-based agent.

        ✅ Use Case: Best for decision-making and reasoning tasks where the agent must break problems down into multiple steps.

        """
        return RunnableMultiActionAgent(
            runnable = create_tool_calling_agent(
                self._llm,
                self.tools,
                prompt=self.prompt,
            ),  # type: ignore
            input_keys_arg=["input"],
            return_keys_arg=["output"],
            **kwargs
        )

    def openai_agent(self, **kwargs):
        """
        Creates OpenAI-like task executor Agent.

        This agent uses reasoning and tool execution iteratively to generate responses.

        Returns:
            RunnableMultiActionAgent: A ReAct-based agent.

        ✅ Use Case: Best for decision-making and reasoning tasks where the agent must break problems down into multiple steps.

        """
        return RunnableMultiActionAgent(
            runnable = create_openai_functions_agent(
                self._llm,
                self.tools,
                prompt=self.prompt
            ),  # type: ignore
            input_keys_arg=["input"],
            return_keys_arg=["output"],
            **kwargs
        )

    def sql_agent(self, dsn: str, **kwargs):
        """
        Creates a SQL Agent.

        This agent is designed to work with SQL queries and databases.

        Returns:
            AgentExecutor: A SQL-based AgentExecutor.

        ✅ Use Case: Best for querying databases and working with SQL data.
        """
        db = SQLDatabase.from_uri(dsn)
        toolkit = SQLDatabaseToolkit(db=db, llm=self._llm)
        # prompt_template = hub.pull("langchain-ai/sql-agent-system-prompt")
        return create_sql_agent(
            llm=self._llm,
            toolkit=toolkit,
            db=db,
            agent_type= "openai-tools",
            extra_tools=self.tools,
            max_iterations=5,
            handle_parsing_errors=True,
            verbose=True,
            prompt=self.prompt,
            agent_executor_kwargs = {"return_intermediate_steps": False}
        )

    def get_executor(
        self,
        agent: RunnableAgent,
        tools: list,
        verbose: bool = True,
        **kwargs
    ):
        """Create a new AgentExecutor.
        """
        return AgentExecutor(
            agent=agent,
            tools=tools,
            verbose=verbose,
            return_intermediate_steps=False,
            max_iterations=5,
            max_execution_time=360,
            handle_parsing_errors=True,
            # memory=self.memory,
            **kwargs,
        )

    def get_agent(self):
        return self.get_executor(self.agent, self.tools)

    async def configure(self, app=None) -> None:
        """Basic Configuration of Agent.
        """
        await super(BasicAgent, self).configure(app)
        # Configure LLM:
        self.configure_llm(use_chat=True)
        # Conversation History:
        self.memory = self.get_memory()
        # 1. Initialize the Agent (as the base for RunnableMultiActionAgent)
        if self.agent_type == 'zero_shot':
            self.agent = self.runnable_agent()
        elif self.agent_type == 'function_calling':
            self.agent = self.function_calling_agent()
        elif self.agent_type == 'openai':
            self.agent = self.openai_agent()
        # elif self.agent_type == 'json':
        #     self.agent = self.runnable_json_agent()
        # elif self.agent_type == 'sql':
        #     self.agent = self.sql_agent()
        else:
            self.agent = self.runnable_agent()
        # self.agent = self.openai_agent()
        # 2. Create Agent Executor - This is where we typically run the agent.
        #  While RunnableMultiActionAgent itself might be "runnable",
        #  we often use AgentExecutor to manage the agent's execution loop.
        self._agent = self.get_executor(self.agent, self.tools)

    async def question(
            self,
            question: str = None,
            **kwargs
    ):
        """question.

        Args:
            question (str): The question to ask the chatbot.
            memory (Any): The memory to use.

        Returns:
            Any: The response from the Agent.

        """
        # TODO: adding the vector-search to the agent
        input_question = {
            "input": question
        }
        result = self._agent.invoke(input_question)
        try:
            response = AgentResponse(question=question, **result)
            # response.response = self.as_markdown(
            #     response
            # )
            return response
        except Exception as e:
            self.logger.exception(
                f"Error on response: {e}"
            )
            raise

    @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(3))
    async def invoke(self, query: str):
        """invoke.

        Args:
            query (str): The query to ask the chatbot.

        Returns:
            str: The response from the chatbot.

        """
        input_question = {
            "input": query
        }
        result = await self._agent.ainvoke(input_question)
        try:
            response = AgentResponse(question=query, **result)
            try:
                return self.as_markdown(
                    response
                ), response
            except Exception as exc:
                self.logger.exception(
                    f"Error on response: {exc}"
                )
                return result.get('output', None), None
        except Exception as e:
            return result, e

    async def __aenter__(self):
        if not self._agent:
            await self.configure()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._agent = None



from abc import ABC, abstractmethod
from ...conf import (
    MAX_BATCH_SIZE,
    EMBEDDING_DEFAULT_MODEL,
    EMBEDDING_DEVICE
)

class AbstractEmbed(ABC):
    """A wrapper class for Create embeddings."""
    model_name: str = EMBEDDING_DEFAULT_MODEL
    encode_kwargs: str = {
        'normalize_embeddings': True,
        "batch_size": MAX_BATCH_SIZE
    }
    model_kwargs = {
        'device': EMBEDDING_DEVICE,
        'trust_remote_code':True
    }

    def __init__(self, model_name: str = None, **kwargs):
        self._embedding = self._create_embedding(model_name, **kwargs)

    @property
    def embedding(self):
        return self._embedding

    def free(self):
        """
        Free the resources.
        """
        pass

    def _get_device(self):
        return EMBEDDING_DEVICE

    @abstractmethod
    def _create_embedding(self, model_name: str = None, **kwargs):
        """
        Create Embedding Model.
        Args:
            model_name (str): The name of the model to use.

        Returns:
            Callable: Embedding Model.
        """
        pass



from aiohttp import web
from navigator_auth.decorators import (
    is_authenticated,
    user_session,
    allowed_organizations
)
from navigator.views import BaseView
from ..bots.abstract import AbstractBot


@is_authenticated()
@user_session()
class ChatHandler(BaseView):
    """
    ChatHandler.
    description: Chat Handler for Parrot Application.
    """

    async def get(self, **kwargs):
        """
        get.
        description: Get method for ChatHandler.
        """
        name = self.request.match_info.get('chatbot_name', None)
        if not name:
            return self.json_response({
                "message": "Welcome to Parrot Chatbot Service."
            })
        else:
            # retrieve chatbof information:
            manager = self.request.app['bot_manager']
            chatbot = manager.get_bot(name)
            if not chatbot:
                return self.error(
                    f"Chatbot {name} not found.",
                    status=404
                )
            config_file = getattr(chatbot, 'config_file', None)
            return self.json_response({
                "chatbot": chatbot.name,
                "description": chatbot.description,
                "role": chatbot.role,
                "embedding_model": chatbot.embedding_model,
                "llm": f"{chatbot.llm!r}",
                "temperature": chatbot.llm.temperature,
                "config_file": config_file
            })

    async def post(self, *args, **kwargs):
        """
        post.
        description: Post method for ChatHandler.

        Use this method to interact with a Chatbot.
        """
        app = self.request.app
        name = self.request.match_info.get('chatbot_name', None)
        qs = self.query_parameters(self.request)
        data = await self.request.json()
        if not 'query' in data:
            return self.json_response(
                {
                "message": "No query was found."
                },
                status=400
            )
        if 'use_llm' in qs:
            # passing another LLM to the Chatbot:
            llm = qs.get('use_llm')
        else:
            llm = None
        try:
            manager = app['bot_manager']
        except KeyError:
            return self.json_response(
                {
                "message": "Chatbot Manager is not installed."
                },
                status=404
            )
        try:
            chatbot: AbstractBot = manager.get_bot(name)
            if not chatbot:
                raise KeyError(
                    f"Chatbot {name} not found."
                )
        except (TypeError, KeyError):
            return self.json_response(
                {
                "message": f"Chatbot {name} not found."
                },
                status=404
            )
        # getting the question:
        question = data.pop('query')
        try:
            session = self.request.session
        except AttributeError:
            session = None
        if not session:
            return self.json_response(
                {
                "message": "User Session is required to interact with a Chatbot."
                },
                status=400
            )
        try:
            async with chatbot.retrieval(request=self.request) as retrieval:
                session_id = session.get('session_id', None)
                memory_key = f'{session.session_id}_{name}_message_store'
                memory = retrieval.get_memory(session_id=memory_key)
                result = await retrieval.invoke(
                    question=question,
                    llm=llm,
                    memory=memory,
                    **data
                )
                # Drop "memory" information:
                result.chat_history = None
                result.source_documents = None
                return self.json_response(response=result)
        except ValueError as exc:
            return self.error(
                f"{exc}",
                exception=exc,
                status=400
            )
        except web.HTTPException as exc:
            return self.error(
                f"{exc}",
                exception=exc,
                status=400
            )
        except Exception as exc:
            return self.error(
                f"Error invoking chatbot {name}: {exc}",
                exception=exc,
                status=400
            )


@is_authenticated()
@user_session()
class BotHandler(BaseView):
    """BotHandler.


    Use this handler to interact with a brand new chatbot, consuming a configuration.
    """
    async def put(self):
        """Create a New Bot (passing a configuration).
        """
        try:
            manager = self.request.app['bot_manager']
        except KeyError:
            return self.json_response(
                {
                "message": "Chatbot Manager is not installed."
                },
                status=404
            )
        # TODO: Making a Validation of data
        data = await self.request.json()
        name = data.pop('name', None)
        if not name:
            return self.json_response(
                {
                "message": "Name for Bot Creation is required."
                },
                status=400
            )
        try:
            chatbot = manager.create_chatbot(name=name, **data)
            await chatbot.configure(name=name, app=self.request.app)
            return self.json_response(
                {
                    "message": f"Chatbot {name} created successfully."
                }
            )
        except Exception as exc:
            return self.error(
                f"Error creating chatbot {name}: {exc}",
                exception=exc,
                status=400
            )



import requests
from crewai_tools import BaseTool


class ValidateURLsTool(BaseTool):
    """Validate URLs Tool."""
    name: str = "Validate URLs"
    description: str = "Validate URLs for status code and response time."

    def _run(self, urls: list) -> dict:
        """Run the Validate URLs Tool."""
        valid_urls = []
        for url in urls:
            try:
                response = requests.head(url, allow_redirects=True)
                if response.status_code == 200:
                    valid_urls.append(url)
            except requests.RequestException:
                # Handle exceptions or log errors here
                continue
        return valid_urls



from collections.abc import Callable
from typing import Union, Optional
from qdrant_client import QdrantClient, models
from langchain.docstore.document import Document
from langchain.memory import VectorStoreRetrieverMemory
from langchain_qdrant import Qdrant
from .abstract import AbstractStore
from ..conf import (
    QDRANT_PROTOCOL,
    QDRANT_HOST,
    QDRANT_PORT,
    QDRANT_USE_HTTPS,
    QDRANT_CONN_TYPE,
    QDRANT_URL
)


class QdrantStore(AbstractStore):
    """Qdrant DB Store Class.

    Using Qdrant as a Document Vector Store.

    """

    def __init__(
        self,
        embedding_model: Union[dict, str] = None,
        embedding: Union[dict, Callable] = None,
        **kwargs
    ):
        super().__init__(
            embedding_model=embedding_model,
            embedding=embedding,
            **kwargs
        )
        self.client: Optional[QdrantClient] = None
        self.host: str = kwargs.pop('host', QDRANT_HOST)
        self.port: int = kwargs.pop('port', QDRANT_PORT)
        self.use_https: bool = kwargs.pop('use_https', QDRANT_USE_HTTPS)
        self.protocol: str = kwargs.pop('protocol', QDRANT_PROTOCOL)
        self.conn_type: str = kwargs.pop('conn_type', QDRANT_CONN_TYPE)
        self.url: str = kwargs.pop('url', QDRANT_URL)
        self._host: str = f"{self.protocol}://{self.host}:{self.port}"

    async def connection(self):
        """Initialize Qdrant vector store.

        Connects to a Qdrant instance, if connection fails, raise error.
        """
        try:
            self._connection = QdrantClient(
                host=self.host,
                port=self.port,
            )
            collections = self._connection.get_collections().collections
            collection_names = [collection.name for collection in collections]
            if self.collection_name not in collection_names:
                print(
                    f"Collection '{self.collection_name}' not found, creating new one."
                )
                self._connection.create_collection(
                    collection_name=self.collection_name,
                    vectors_config=models.VectorParams(
                        size=len(self._embed_.embedding.embed_query("test")), distance=models.Distance.EUCLID
                    )
                )
            self.client = Qdrant(
                client=self._connection,
                collection_name=self.collection_name,
                embedding_function=self._embed_.embedding
            )
        except Exception as e:
            print(f"Error connecting to Qdrant: {e}")
            raise
        self._connected = True
        return self._connection

    async def disconnect(self) -> None:
        """Disconnects from Qdrant instance."""
        self.client = None # No explicit disconnect needed for QdrantClient in this context.
        self._connection = None
        self._connected = False

    def get_vector(
        self,
        embedding: Optional[Callable] = None,
    ) -> Qdrant:
        """Returns Qdrant VectorStore instance."""
        if embedding is not None:
            _embed_ = embedding
        else:
            _embed_ = self.create_embedding(
                embedding_model=self.embedding_model
            )
        if not self._connection:
            # Re-establish client if lost, but should not happen frequently in well-managed context.
            self._connection = QdrantClient(host=self.host, port=self.port)
        return Qdrant(
            client=self._connection,
            collection_name=self.collection_name,
            embedding_function=_embed_
        )

    async def from_documents(self, documents: list[Document], **kwargs):
        """Save Documents as Vectors in Qdrant."""
        vectordb = await Qdrant.afrom_documents(
            documents=documents,
            embedding=self._embed_.embedding,
            client=self.client,
            collection_name=self.collection_name,
            **kwargs
        )
        return vectordb

    async def add_documents(
        self,
        documents: list[Document],
        embedding: Optional[Callable] = None,
    ) -> bool:
        """Add Documents to Qdrant."""
        async with self:
            vector_db = self.get_vector(embedding=embedding)
            await vector_db.aadd_documents(documents=documents)
        return True

    async def update_documents(
        self,
        documents: list[Document],
        embedding: Optional[Callable] = None,
    ) -> bool:
        """Update Documents in Qdrant."""
        async with self:
            vector_db = self.get_vector(embedding=embedding)
            if all('id' in doc for doc in documents):
                ids = [doc.pop('id') for doc in documents]
                vector_db.delete(ids=ids)  # Remove old entries
                await vector_db.aadd_documents(documents=documents)  # Add new versions
                return True
            return False

    async def similarity_search(
        self,
        query: str,
        embedding: Optional[Callable] = None,
        limit: int = 2,
    ) -> list:
        """Performs similarity search in Qdrant."""
        async with self:
            vector_db = self.get_vector(embedding=embedding)
            return await vector_db.asimilarity_search(query, k=limit) # Use asimilarity_search for async

    def memory_retriever(
        self,
        documents: Optional[list] = None,
        num_results: int = 5
    ) -> VectorStoreRetrieverMemory:
        """Retrieves stored memory-based documents."""
        if not documents:
            documents = []
        vectordb = Qdrant.from_documents(
            documents=documents,
            embedding=self._embed_.embedding,
            client=self.client,
            collection_name=self.collection_name
        )
        retriever = Qdrant.as_retriever(
            vectordb,
            search_kwargs=dict(k=num_results)
        )
        return VectorStoreRetrieverMemory(retriever=retriever)



supported_embeddings = {
    'openai': 'OpenAIEmbed',
    'google': 'GoogleEmbed',
    'vertexai': 'VertexAIEmbed',
    'huggingface': 'HugginfaceEmbed',
    'fastembed': 'FastembedEmbed',
    'bge': 'BgeEmbed',
    'ollama': 'OllamaEmbed',
    'transformers': 'TransformersEmbed'
}



from abc import ABC, abstractmethod
from typing import List, Union
import importlib
from collections.abc import Callable
from langchain.docstore.document import Document
from langchain_core.vectorstores import VectorStoreRetriever
from navconfig.logging import logging
from ..conf import (
    EMBEDDING_DEFAULT_MODEL
)
from ..exceptions import ConfigError  # pylint: disable=E0611
from .embeddings import supported_embeddings


class AbstractStore(ABC):
    """AbstractStore class.

        Base class for all Database Vector Stores.
    Args:
        embeddings (str): Embedding name.

    Supported Vector Stores:
        - Qdrant
        - Milvus
        - Faiss
        - Chroma
        - PgVector
    """

    def __init__(
        self,
        embedding_model: Union[dict, str] = None,
        embedding: Union[dict, Callable] = None,
        **kwargs
    ):
        self.client: Callable = None
        self.vector: Callable = None
        self._embed_: Callable = None
        self._connected: bool = False
        if embedding_model is not None:
            if isinstance(embedding_model, str):
                self.embedding_model = {
                    'model_name': embedding_model,
                    'model_type': 'huggingface'
                }
            elif isinstance(embedding_model, dict):
                self.embedding_model = embedding_model
            else:
                raise ValueError(
                    "Embedding Model must be a string or a dictionary."
                )
        # Use or not connection to a vector database:
        self._use_database: bool = kwargs.get('use_database', True)
        # Database Information:
        self.collection_name: str = kwargs.get('collection_name', 'my_collection')
        self.dimension: int = kwargs.get("dimension", 768)
        self._metric_type: str = kwargs.get("metric_type", 'COSINE')
        self._index_type: str = kwargs.get("index_type", 'IVF_FLAT')
        self.database: str = kwargs.get('database', '')
        self.index_name = kwargs.get("index_name", "my_index")
        if embedding is not None:
            if isinstance(embedding, str):
                self.embedding_model = {
                    'model_name': embedding,
                    'model_type': 'huggingface'
                }
            elif isinstance(embedding, dict):
                self.embedding_model = embedding
            else:
                # is a callable:
                self.embedding_model = {
                    'model_name': EMBEDDING_DEFAULT_MODEL,
                    'model_type': 'huggingface'
                }
                self._embed_ = embedding
        self.logger = logging.getLogger(
            f"Store.{__name__}"
        )
        # Client Connection (if required):
        self._connection = None
        # Create the Embedding Model:
        self._embed_ = self.create_embedding(
            embedding_model=self.embedding_model
        )

    @property
    def connected(self) -> bool:
        return self._connected

    def is_connected(self):
        return self._connected

    @abstractmethod
    async def connection(self) -> tuple:
        pass

    @abstractmethod
    async def disconnect(self) -> None:
        pass

    # Async Context Manager
    async def __aenter__(self):
        if self._use_database:
            if not self._connection:
                await self.connection()
        return self

    async def _free_resources(self):
        self._embed_.free()
        self._embed_ = None

    async def __aexit__(self, exc_type, exc_value, traceback):
        # closing Embedding
        try:
            await self.disconnect()
        except RuntimeError:
            pass

    @abstractmethod
    def get_vector(self, metric_type: str = None, **kwargs):
        pass

    def get_vectorstore(self):
        return self.get_vector()

    @abstractmethod
    async def similarity_search(
        self,
        query: str,
        collection: Union[str, None] = None,
        limit: int = 2
    ) -> list:  # noqa
        pass

    @abstractmethod
    async def from_documents(
        self,
        documents: List[Document],
        collection: Union[str, None] = None,
        **kwargs
    ) -> Callable:
        """
        Create Vector Store from Documents.

        Args:
            documents (List[Document]): List of Documents.
            collection (str): Collection Name.
            kwargs: Additional Arguments.

        Returns:
            Callable VectorStore.
        """

    @abstractmethod
    async def add_documents(
        self,
        documents: List[Document],
        collection: Union[str, None] = None,
        **kwargs
    ) -> None:
        """
        Add Documents to Vector Store.

        Args:
            documents (List[Document]): List of Documents.
            collection (str): Collection Name.
            kwargs: Additional Arguments.

        Returns:
            None.
        """

    def create_embedding(
        self,

        embedding_model: dict,
        **kwargs
    ):
        """
        Create Embedding Model.

        Args:
            embedding_model (dict): Embedding Model Configuration.
            kwargs: Additional Arguments.

        Returns:
            Callable: Embedding Model.

        """
        model_type = embedding_model.get('model_type', 'huggingface')
        model_name = embedding_model.get('model_name', EMBEDDING_DEFAULT_MODEL)
        if model_type not in supported_embeddings:
            raise ConfigError(
                f"Embedding Model Type: {model_type} not supported."
            )
        embed_cls = supported_embeddings[model_type]
        cls_path = f".embeddings.{model_type}"  # Relative module path
        try:
            embed_module = importlib.import_module(
                cls_path,
                package=__package__
            )
            embed_obj = getattr(embed_module, embed_cls)
            return embed_obj(
                model_name=model_name,
                **kwargs
            )
        except ImportError as e:
            raise ConfigError(
                f"Error Importing Embedding Model: {model_type}"
            ) from e

    def get_default_embedding(self):
        embed_model = {
            'model_name': EMBEDDING_DEFAULT_MODEL,
            'model_type': 'huggingface'
        }
        return self.create_embedding(
            embedding_model=embed_model
        )

    def generate_embedding(self, documents: List[Document]):
        if not self._embed_:
            self._embed_ = self.get_default_embedding()

        # Using the Embed Model to Generate Embeddings:
        embeddings = self._embed_.embed_documents(documents)
        return embeddings

    def as_retriever(
        self,
        metric_type: str = 'COSINE',
        index_type: str = 'IVF_FLAT',
        search_type: str = 'similarity',
        chain_type: str = 'stuff',
        search_kwargs: dict = None
    ) -> Callable:
        vector = self.get_vector(metric_type=metric_type, index_type=index_type)
        if not vector:
            raise ConfigError(
                "Vector Store is not connected. Check your connection."
            )
        return VectorStoreRetriever(
            vectorstore=vector,
            search_type=search_type,
            chain_type=chain_type,
            search_kwargs=search_kwargs
        )



from typing import List, Dict, Any, Union
import random
from orjson import JSONDecodeError
from pydantic import BaseModel, Field
# from pydantic.v1 import BaseModel, Field
from langchain_core.tools import BaseTool, BaseToolkit, StructuredTool, ToolException, Tool
from datamodel.parsers.json import json_decoder, json_encoder  # pylint: disable=E0611
from datamodel.exceptions import ParserError  # pylint: disable=E0611
from navconfig import config
from ..interfaces.http import HTTPService, ua


ctt_list: list = [
    "f3dbf688e45146555bb2b8604a993601",
    "06f4dfe367e87866397ef32302f5042e",
    "4e07e03ff03f5debc4e09ac4db9239ac"
]

sid_list: list = [
    "d4fa1142-2998-4b68-af78-46d821bb3e1f",
    "9627390e-b423-459f-83ee-7964dd05c9a8"
]

class BestBuyProductAvailabilityInput(BaseModel):
    """Input for the BestBuy product availability tool."""
    zipcode: str = Field(..., description="The ZIP code to check availability in")
    sku: str = Field(..., description="The SKU of the product to check")
    location_id: str = Field(
        ..., description="Optional specific location ID to check"
    )
    show_only_in_stock: bool = Field(
        False, description="Whether to only show stores with product in stock"
    )

    model_config = {
        "arbitrary_types_allowed": True,
        "extra": "forbid",  # Helps with compatibility
        "json_schema_extra": {
            "required": ["zipcode", "sku", "location_id"]
        }
    }

class BestBuyToolkit(BaseToolkit):
    """Toolkit for interacting with BestBuy's API."""

    api_key: str = Field(default=config.get('BESTBUY_APIKEY'))

    http_service: HTTPService = Field(default_factory=lambda: HTTPService(
        use_proxy=True,
        cookies={
            "CTT": random.choice(ctt_list),
            "SID": random.choice(sid_list),
            "bby_rdp": "l",
            "bm_sz": "9F5ED0110AF18594E2347A89BB4AB998~YAAQxm1lX6EqYHGSAQAAw+apmhkhXIeGYEc4KnzUMsjeac3xEoQmTNz5+of62i3RXQL6fUI+0FvCb/jgSjiVQOcfaSF+LdLkOXP1F4urgeIcqp/dBAhu5MvZXaCQsT06bwr7j21ozhFfTTWhjz1HmZN8wecsE6WGbK6wXp/33ODKlLaGWkTutqHbkzvMiiHXBCs9hT8jVny0REfita4AfqTK85Y6/M6Uq4IaDLPBLnTtJ0cTlPHk1HmkG5EsnI46llghcx1KZnCGnvZfHdb2ME9YZJ2GmC2b7dNmAgyL/gSVpoNdCJOj5Jk6z/MCVhZ81OZfX4S01E2F1mBGq4uV5/1oK2KR4YgZP4dsTN8izEEPybUKGY3CyM1gOUc=~3556420~4277810",
            "bby_cbc_lb": "p-browse-e",
            "intl_splash": "false"
        },
        headers={
            "Host": "www.bestbuy.com",
            "Referer": "https://www.bestbuy.com/",
            "TE": "trailers",
            "Accept-Language": "en-US,en;q=0.5",
        }
    ))

    # Add this model_config to allow arbitrary types
    model_config = {
        "arbitrary_types_allowed": True
    }

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
            self._get_availability_tool(),
            self._get_product_information()
        ]

    def _get_product_information(self) -> StructuredTool:
        """Create a tool for getting product information from BestBuy."""

        async def _product_information(tool_input: Union[str, dict]) -> List[dict]:
            """Get product information from BestBuy based on SKU, name, or search terms."""
            # https://api.bestbuy.com/v1/products(name={product_name})?format=json&show=sku,name,salePrice,customerReviewAverage,customerReviewCount,manufacturer,modelNumber&apiKey={api_key}
            if isinstance(tool_input, dict):
                # Direct call with dict
                input_data = tool_input
            elif isinstance(tool_input, str):
                # Might be a JSON string from LLM
                try:
                    # input_data = json.loads(tool_input)
                    input_data = json_decoder(tool_input)
                except (ParserError, JSONDecodeError):
                    # Not valid JSON, treat as zipcode with missing other params
                    input_data = {"product_name": tool_input}
            else:
                # Some other type that we don't expect
                input_data = {"product_name": str(tool_input)}

            # Product Name to be used in the search:
            print('PRODUCT > ', input_data)

            product_name = input_data.get("product_name", None)
            search_terms = input_data.get("search_terms", None)
            print(search_terms)
            if search_terms:
                search_terms = search_terms.split(',')
                # I need in format: search=oven&search=stainless&search=steel
                search_terms = '&'.join([f"search={term.strip()}" for term in search_terms])
                url = f"https://api.bestbuy.com/v1/products({search_terms})?format=json&show=sku,name,salePrice,customerReviewAverage,customerReviewCount,manufacturer,modelNumber&apiKey={self.api_key}"
            elif product_name:
                if ',' in product_name:
                    search_terms = product_name.split(',')
                    search_terms = '&'.join([f"search={term.strip()}" for term in search_terms])
                else:
                    search_terms = f"name={product_name.strip()}"
                url = f"https://api.bestbuy.com/v1/products({search_terms})?format=json&show=sku,name,salePrice,customerReviewAverage,customerReviewCount,manufacturer,modelNumber&apiKey={self.api_key}"
            else:
                raise ToolException(
                    "Either product_name or search_terms must be provided."
                )
            # URL for BestBuy's Product API
            result, error = await self.http_service._request(
                url=url,
                method="GET",
                use_json=True,
                use_proxy=True,
                headers={
                    "User-Agent": random.choice(ua)
                },
                use_ssl=True,
                follow_redirects=True,
                raise_for_status=True,
                full_response=False
            )
            products = result.get('products', [])
            if not products:
                return "No products found."

            return products


        return StructuredTool.from_function(
            name="product_information",
            description=(
                "Use this tool to search for product information. "
                "Input must be the search terms: "
                "- search terms: different words separated by commas "
                " Example input: {{\"search_terms\": \"oven,stainless\"}}"
            ),
            func=_product_information,
            coroutine=_product_information,
            infer_schema=False,
            return_direct=False,
            handle_tool_error=True
        )

    def _get_availability_tool(self) -> StructuredTool:
        """Create a tool for checking product availability at BestBuy."""

        async def _check_availability(tool_input: Union[str, dict]) -> str:
            """Check BestBuy product availability based on SKU and location."""

            # Input validation
            # tool_input: str
            try:
                if isinstance(tool_input, dict):
                    # Direct call with dict
                    input_data = tool_input
                elif isinstance(tool_input, str):
                    # Might be a JSON string from LLM
                    try:
                        # input_data = json.loads(tool_input)
                        input_data = json_decoder(tool_input)
                    except (ParserError, JSONDecodeError):
                        # Not valid JSON, treat as zipcode with missing other params
                        input_data = {"zipcode": tool_input}
                else:
                    # Some other type that we don't expect
                    input_data = {"zipcode": str(tool_input)}
                # Extract fields
                zipcode = input_data.get("zipcode")
                sku = input_data.get("sku")
                location_id = input_data.get("location_id")
                show_only_in_stock = input_data.get("show_only_in_stock", False)

            except Exception as e:
                # Final fallback - treat the entire input as a zipcode
                zipcode = str(tool_input)
                sku = None
                location_id = None
                show_only_in_stock = False

            # Input validation
            if not zipcode:
                raise ToolException("ZIP code is required to check product availability")
            if not sku:
                raise ToolException("Product SKU is required to check availability")
            if not location_id:
                raise ToolException("Store location ID is required to check availability")

            # Prepare the payload for the API call
            payload = {
                "locationId": location_id,
                "zipCode": zipcode,
                "showOnShelf": True,
                "lookupInStoreQuantity": True,
                "xboxAllAccess": False,
                "consolidated": True,
                "showOnlyOnShelf": False,
                "showInStore": True,
                "pickupTypes": [
                    "UPS_ACCESS_POINT",
                    "FEDEX_HAL"
                ],
                "onlyBestBuyLocations": True,
                # TODO: add more products
                "items": [
                    {
                        "sku": sku,
                        "condition": None,
                        "quantity": 1,
                        "itemSeqNumber": "1",
                        "reservationToken": None,
                        "selectedServices": [],
                        "requiredAccessories": [],
                        "isTradeIn": False,
                        "isLeased": False
                    }
                ]
            }

            # Make the API call using the HTTP service
            try:
                # URL for BestBuy's availability API
                url = "https://www.bestbuy.com/productfulfillment/c/api/2.0/storeAvailability"

                # Make POST request with JSON payload
                result, error = await self.http_service._request(
                    url=url,
                    method="POST",
                    data=payload,
                    use_json=True,
                    use_proxy=True,
                    headers={
                        "User-Agent": random.choice(ua)
                    },
                    use_ssl=True,
                    follow_redirects=True,
                    raise_for_status=True,
                    full_response=False
                )

                if error:
                    raise ToolException(
                        f"Error checking BestBuy availability: {error}"
                    )

                # Process and format the response
                if not result:
                    return "No data was returned from BestBuy. The service may be unavailable."

                # Extract relevant information from the result
                formatted_result = self._format_availability_response(result, location_id, sku, show_only_in_stock)
                return formatted_result

            except Exception as e:
                raise ToolException(f"Failed to check BestBuy product availability: {str(e)}")

        return StructuredTool.from_function(
            name="availability",
            description=(
                "Use this tool to check product availability at a specific Best Buy store or zipcode. "
                "Input must be a JSON object with the following fields: "
                "- zipcode (required): The ZIP code to check availability in. "
                "- sku (required): The SKU of the product to check. "
                "- location_id (required): The specific store location ID to check. "
                "- show_only_in_stock (optional): Whether to only show stores with product in stock. "
                "Example input: {{\"zipcode\": \"33928\", \"sku\": \"6428376\", \"location_id\": \"767\", \"show_only_in_stock\": false}}"
            ),
            func=_check_availability,
            # args_schema=BestBuyProductAvailabilityInput,
            coroutine=_check_availability,
            infer_schema=False,
            return_direct=False,
            handle_tool_error=True
        )

    def _format_availability_response(
        self, result: Dict[str, Any], location_id: str, sku: str, show_only_in_stock: bool = False
    ) -> str:
        try:
            # Extract store information from the "ispu" locations
            locations = result.get("ispu", {}).get("locations", [])
            # Match on the "id" key rather than "locationId"
            store = next((loc for loc in locations if loc.get("id") == location_id), None)
            if not store:
                return "No matching store location found."

            store_name = store.get("name", "N/A")
            store_address = store.get("address", "N/A")
            store_city = store.get("city", "N/A")
            store_zip = store.get("zipCode", "N/A")
            store_state = store.get("state", "N/A")
            store_lat = store.get("latitude", "N/A")
            store_long = store.get("longitude", "N/A")

            # Format store hours if available
            open_times = store.get("openTimesMap", {})
            store_hours_str = (
                "\n".join(f"{day}: {hours}" for day, hours in open_times.items())
                if open_times else "N/A"
            )

            # Extract product availability from "ispu" items
            items = result.get("ispu", {}).get("items", [])
            item = next((it for it in items if it.get("sku") == sku), None)
            if not item:
                return "No matching product found."

            print('ITEM > ', item)

            product_instore = item.get("inStoreAvailable", False)
            product_pickup = item.get("pickupEligible", False)

            # Extract the location-specific availability from the item's "locations" list
            item_locations = item.get("locations", [])
            availability = next(
                (loc for loc in item_locations if loc.get("locationId") == location_id), None
            )
            if not availability:
                return "No matching product availability found."

            on_shelf = availability.get("onShelfDisplay", False)
            in_store_quantity = availability.get("inStoreAvailability", {}).get("availableInStoreQuantity", 0)
            available_from = availability.get("inStoreAvailability", {}).get("minDate")

            # Build and return a formatted result string
            result_str = (
                f"Store Name: {store_name}\n"
                f"Address: {store_address}\n"
                f"City: {store_city}\n"
                f"Zip Code: {store_zip}\n"
                f"State: {store_state}\n"
                f"Latitude: {store_lat}\n"
                f"Longitude: {store_long}\n"
                f"Hours:\n{store_hours_str}\n"
                "--------------------------------\n"
                f"Product SKU: {sku}\n"
                f"In-store Available: {product_instore}\n"
                f"Pickup Eligible: {product_pickup}\n"
                f"On Shelf Display: {on_shelf}\n"
                f"Available Quantity: {in_store_quantity}\n"
                f"Available From: {available_from}\n"
                "--------------------------------\n"
            )
            return result_str

        except Exception as e:
            return f"Error formatting availability response: {str(e)}"



"""
Powerful PostgreSQL Vector Database Store with Custom Table Support.
"""
from typing import (
    Any,
    Dict,
    List,
    Tuple,
    Union,
    Optional,
    Sequence
)
from collections.abc import Callable
import asyncio
import uuid
# SQL Alchemy
import sqlalchemy
from sqlalchemy import inspect, text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, ARRAY, Float, JSON, text, func
from sqlalchemy.dialects.postgresql import JSON, JSONB, JSONPATH, UUID, insert
from sqlalchemy.ext.asyncio import create_async_engine, AsyncEngine
from sqlalchemy.future import select
from sqlalchemy.ext.asyncio import AsyncSession
# PgVector
from pgvector.sqlalchemy import Vector  # type: ignore
# Langchain
from langchain_core.embeddings import Embeddings
from langchain.docstore.document import Document
from langchain.memory import VectorStoreRetrieverMemory
from langchain_community.vectorstores.pgembedding import PGEmbedding
from langchain_community.vectorstores.utils import DistanceStrategy
from langchain_postgres.vectorstores import (
    PGVector,
    _get_embedding_collection_store,
    _results_to_docs
)
from datamodel.parsers.json import json_encoder  # pylint: disable=E0611
from .abstract import AbstractStore



Base = declarative_base()


# Define the async classmethods to be attached to our ORM model.
async def aget_by_name(cls, session: AsyncSession, name: str) -> Optional["CustomEmbeddingStore"]:
    # result = await session.execute(select(cls).where(cls.name == name))
    # return result.scalars().first()
    return cls(cmetadata={})


class PgVector(PGVector):
    """
    PgVector extends PGVector so that it uses an existing table from a specified schema.

    When instantiating, you provide:
    - connection: an AsyncEngine (or synchronous engine) to your PostgreSQL database.
    - schema: the database schema where your table lives.
    - table_name: the name of the table that stores the embeddings.
    - embedding_length: the dimension of the embedding vectors.
    - embeddings: your embedding function/model (which must provide embed_query).

    This implementation overrides the _get_embedding_collection_store method to return a tuple of
    ORM model classes that both refer to your table. It validates (using SQLAlchemy’s inspector)
    that the table contains the required columns: 'id', 'embedding', 'document', and 'cmetadata'.

    The returned ORM models can then be used by PGVector’s built-in similarity search and retriever.
    """
    def __init__(
        self,
        embeddings: Embeddings,
        *,
        table_name: str = None,
        schema: str = 'public',
        collection_name: Optional[str] = None,
        id_column: str = 'id',
        **kwargs
    ) -> None:
        self.table_name = table_name
        self.schema = schema
        self._id_column: str = id_column
        self._schema_based: bool = False
        if self.table_name:
            self._schema_based: bool = True
        elif '.' in collection_name:
            self.schema, self.table_name = collection_name.split('.')
            self._schema_based: bool = True
        super().__init__(
            embeddings=embeddings,
            collection_name=collection_name,
            **kwargs
        )

    async def _get_embedding_collection_store(
        self,
        table: str,
        schema: str,
        dimension: int = 768,
        **kwargs
    ) -> Tuple[type, type]:
        """
        Return custom ORM model classes (EmbeddingStore, CollectionStore)
        that both reference the same table.

        In this custom implementation, both the "collection" and "embedding" stores
        are represented by a single table.
        The table is expected to have the following columns:
        - id: unique identifier (String)
        - embedding: the vector column (Vector(dimension))
        - document: text column containing the document
        - cmetadata: JSONB column for metadata

        Raises an error if the table does not have the required schema.
        """
        # Dynamically create the model class.
        attrs = {
            '__tablename__': table,
            '__table_args__': {"schema": schema},
            self._id_column: sqlalchemy.Column(
                sqlalchemy.String,
                primary_key=True,
                index=True,
                unique=True,
                default=lambda: str(uuid.uuid4())
            ),
            'embedding': sqlalchemy.Column(Vector(dimension)),
            'document': sqlalchemy.Column(sqlalchemy.String, nullable=True),
            'cmetadata': sqlalchemy.Column(JSONB, nullable=True),
            # Attach the async classmethods.
            'aget_by_name': classmethod(aget_by_name),
            # 'aget_or_create': classmethod(aget_or_create)
        }
        EmbeddingStore = type("CustomEmbeddingStore", (Base,), attrs)
        EmbeddingStore.__name__ = "EmbeddingStore"
        return (EmbeddingStore, EmbeddingStore)

    async def __apost_init__(
        self,
    ) -> None:
        """Async initialize the store (use lazy approach)."""
        if self._async_init:  # Warning: possible race condition
            return
        self._async_init = True
        if self._schema_based:
            ebstore, cstore = await self._get_embedding_collection_store(
                table=self.table_name,
                schema=self.schema,
                dimension=self._embedding_length
            )
        else:
            ebstore, cstore = _get_embedding_collection_store(
                self._embedding_length
            )
        self.CollectionStore = cstore
        self.EmbeddingStore = ebstore

        if not self._schema_based:
            await self.acreate_tables_if_not_exists()
            await self.acreate_collection()

    async def asimilarity_search(
        self,
        query: str,
        k: int = 4,
        score_threshold: Optional[float] = None,
        filter: Optional[dict] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Run similarity search with PGVector with distance.

        Args:
            query (str): Query text to search for.
            k (int): Number of results to return. Defaults to 4.
            filter (Optional[Dict[str, str]]): Filter by metadata. Defaults to None.

        Returns:
            List of Documents most similar to the query.
        """
        await self.__apost_init__()  # Lazy async init
        embedding = await self.embeddings.aembed_query(query)
        return await self.asimilarity_search_by_vector(
            embedding=embedding,
            k=k,
            score_threshold=score_threshold,
            filter=filter,
        )

    async def asimilarity_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        score_threshold: Optional[float] = None,
        filter: Optional[dict] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Return docs most similar to embedding vector.

        Args:
            embedding: Embedding to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter (Optional[Dict[str, str]]): Filter by metadata. Defaults to None.

        Returns:
            List of Documents most similar to the query vector.
        """
        assert self._async_engine, "This method must be called with async_mode"
        await self.__apost_init__()  # Lazy async init
        docs_and_scores = await self.asimilarity_search_with_score_by_vector(
            embedding=embedding, k=k, score_threshold=score_threshold, filter=filter
        )
        return _results_to_docs(docs_and_scores)

    async def asimilarity_search_with_score_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        score_threshold: Optional[float] = None,
        filter: Optional[dict] = None,
    ) -> List[Tuple[Document, float]]:
        await self.__apost_init__()  # Lazy async init
        async with self._make_async_session() as session:  # type: ignore[arg-type]
            results = await self._aquery_collection(
                session=session, embedding=embedding, k=k, score_threshold=score_threshold, filter=filter
            )
            return self._results_to_docs_and_scores(results)

    async def _aquery_collection(
        self,
        session: AsyncSession,
        embedding: List[float],
        k: int = 4,
        score_threshold: Optional[float] = None,
        filter: Optional[Dict[str, str]] = None,
    ) -> List[Tuple[Document, float]]:
        """Search for similar documents in the collection.

        If score_threshold is provided, returns all documents whose computed distance is below that threshold.
        Otherwise, if k is provided, returns at most k documents.
        """
        async with self._make_async_session() as session:  # type: ignore[arg-type]
            filter_by = []
            if filter:
                if self.use_jsonb:
                    filter_clause = self._create_filter_clause(filter)
                    if filter_clause is not None:
                        filter_by.append(filter_clause)
                else:
                    # For non-JSONB cases, you might use a deprecated method:
                    filter_clauses = self._create_filter_clause_json_deprecated(filter)
                    filter_by.extend(filter_clauses)

            # Compute the distance expression
            distance_expr = self.distance_strategy(embedding).label("distance")
            stmt = (
                sqlalchemy.select(
                    self.EmbeddingStore,
                    self.distance_strategy(embedding).label("distance")
                )
                .filter(*filter_by)
            )
            # If a score threshold is provided, add a filter on the distance.
            if score_threshold is not None:
                stmt = stmt.filter(distance_expr < score_threshold)
            else:
                # Otherwise, limit the number of results.
                stmt = stmt.order_by(distance_expr).limit(k)

            stmt = stmt.order_by(sqlalchemy.asc(distance_expr))
            # Execute the query and return the results.
            results: Sequence[Any] = (await session.execute(stmt)).all()
            return results

    def _results_to_docs_and_scores(self, results: Any) -> List[Tuple[Document, float]]:
        """Return docs and scores from results."""
        id_col = getattr(self, "_id_column", "id")
        docs = [
            (
                Document(
                    id=str(getattr(result.EmbeddingStore, id_col)),
                    page_content=result.EmbeddingStore.document,
                    metadata=result.EmbeddingStore.cmetadata,
                ),
                result.distance if self.embeddings is not None else None,
            )
            for result in results
        ]
        return docs


class PgvectorStore(AbstractStore):
    """Pgvector Store Class.

    Using PostgreSQL + PgVector to saving vectors in database.
    """
    def __init__(
        self,
        embedding_model: Union[dict, str] = None,
        embedding: Union[dict, Callable] = None,
        **kwargs
    ):
        super().__init__(
            embedding_model=embedding_model,
            embedding=embedding,
            **kwargs
        )
        self.table: str = kwargs.get('table', None)
        self.schema: str = kwargs.get('schema', 'public')
        self._id_column = kwargs.get('id_column', 'id')
        if self.table and not self.collection_name:
            self.collection_name = f"{self.schema}.{self.table}"
        self.dsn = kwargs.get('dsn', self.database)
        self._drop: bool = kwargs.pop('drop', False)
        self._connection: AsyncEngine = None

    async def collection_exists(self, collection: str = None) -> bool:
        """Check if a collection exists in the database."""
        if not collection:
            collection = self.collection_name
        async with self._connection.connect() as conn:
            # ✅ Check if the collection (table) exists
            check_query = f"""
            SELECT EXISTS (
                SELECT FROM information_schema.tables
                WHERE table_name = '{collection}'
            );
            """
            result = await conn.execute(sqlalchemy.text(check_query))
            return bool(result.scalar())

    async def connection(self, alias: str = None):
        """Connection to DuckDB.

        Args:
            alias (str): Database alias.

        Returns:
            Callable: DuckDB connection.

        """
        self._connection = create_async_engine(self.dsn, future=True, echo=False)
        async with self._connection.begin() as conn:
            if getattr(self, "_drop", False):
                vectorstore = PgVector(
                    embeddings=self._embed_.embedding,
                    table_name=self.table,
                    schema=self.schema,
                    id_column=self._id_column,
                    collection_name=self.collection_name,
                    embedding_length=self.dimension,
                    connection=self._connection,
                    use_jsonb=True,
                    create_extension=False
                )
                await vectorstore.adrop_tables()
            if not await self.collection_exists(self.collection_name):
                print(f"⚠️ Collection `{self.collection_name}` not found. Creating a new one...")
                await self.create_collection(self.collection_name)
        self._connected = True
        return self._connection

    def engine(self):
        return self._connection

    async def disconnect(self) -> None:
        """
        Closing the Connection on DuckDB
        """
        try:
            if self._connection:
                await self._connection.dispose()
        except Exception as err:
            raise RuntimeError(
                message=f"{__name__!s}: Closing Error: {err!s}"
            ) from err
        finally:
            self._connection = None
            self._connected = False

    async def create_collection(self, collection: str) -> None:
        """Create a new collection in the database."""
        async with self._connection.connect() as conn:
            # ✅ Create the collection in PgVector
            _embed_ = self._embed_ or self.create_embedding(
                embedding_model=self.embedding_model
            )
            self._client = PgVector(
                embeddings=_embed_.embedding,
                embedding_length=self.dimension,
                collection_name=self.collection_name,
                connection=self._connection,
                use_jsonb=True,
                create_extension=False
            )
            print(
                f"✅ Collection `{self.collection_name}` created successfully."
            )

    def get_vector(
        self,
        table: Optional[str] = None,
        schema: Optional[str] = None,
        collection: Union[str, None] = None,
        embedding: Optional[Callable] = None,
        **kwargs
    ) -> PGVector:
        """
        This function retrieves a vector from the specified collection using the provided embedding.
        If no collection is specified, it uses the default collection name.
        If no embedding is provided, it creates a new embedding using the specified embedding model.

        Parameters:
        - collection (Union[str, None]): The name of the collection from which to retrieve the vector.
        - embedding (Optional[Callable]): The embedding function to use for vector retrieval.
        - kwargs: Additional keyword arguments to pass to the PGVector constructor.

        Returns:
        - PGVector: The retrieved vector from the specified collection.
        """
        if not table:
            table = self.table
        if not schema:
            schema = self.schema
        if not collection:
            collection = self.collection_name
        if embedding is not None:
            _embed_ = embedding
        else:
            _embed_ = self.create_embedding(
                embedding_model=self.embedding_model
            )
        return PgVector(
            connection=self._connection,
            table_name=table,
            schema=schema,
            id_column=self._id_column,
            collection_name=collection,
            embedding_length=self.dimension,
            embeddings=_embed_.embedding,
            logger=self.logger,
            async_mode=True,
            use_jsonb=True,
            create_extension=False,
            **kwargs
        )

    def memory_retriever(
        self,
        documents: Optional[List[Document]] = None,
        num_results: int  = 5
    ) -> VectorStoreRetrieverMemory:
        _embed_ = self._embed_ or self.create_embedding(
                embedding_model=self.embedding_model
        )
        vectordb = PgVector.from_documents(
            documents or [],
            embedding=_embed_.embedding,
            connection=self._connection,
            collection_name=self.collection_name,
            embedding_length=self.dimension,
            use_jsonb=True,
            async_mode=True,
            create_extension=False,
        )
        retriever = PgVector.as_retriever(
            vectordb,
            search_kwargs=dict(k=num_results)
        )
        return VectorStoreRetrieverMemory(retriever=retriever)

    async def from_documents(
        self,
        documents: List[Document],
        table: Optional[str] = None,
        schema: Optional[str] = 'public',
        collection: Union[str, None] = None,
        **kwargs
    ) -> None:
        """Save Documents as Vectors in VectorStore."""
        _embed_ = self._embed_ or self.create_embedding(
                embedding_model=self.embedding_model
        )
        if not collection:
            collection = self.collection_name
        vectordb = await PgVector.afrom_documents(
            documents,
            connection=self._connection,
            table_name=table,
            schema=schema,
            id_column=self._id_column,
            collection_name=collection,
            embedding=_embed_.embedding,
            embedding_length=self.dimension,
            use_jsonb=True,
            async_mode=True,
        )
        return vectordb

    async def add_documents(
        self,
        documents: List[Document],
        collection: Union[str, None] = None,
        **kwargs
    ) -> None:
        """Save Documents as Vectors in VectorStore."""
        if not collection:
            collection = self.collection_name
        vectordb = self.get_vector(collection=collection, **kwargs)
        # Asynchronously add documents to PGVector
        await vectordb.aadd_documents(documents)

    async def similarity_search(
        self,
        query: str,
        table: Optional[str] = None,
        schema: Optional[str] = None,
        collection: Union[str, None] = None,
        limit: int = 2,
        score_threshold: Optional[float] = None,
        filter: Optional[dict] = None,
        **kwargs
    ) -> List[Document]:
        """Search for similar documents in VectorStore."""
        if not table:
            table = self.table
        if not schema:
            schema = self.schema
        if collection is None:
            collection = self.collection_name
        async with self:
            vector_db = self.get_vector(table=table, schema=schema, collection=collection, **kwargs)
            return await vector_db.asimilarity_search(
                query,
                k=limit,
                score_threshold=score_threshold,
                filter=filter
            )

    async def create_embedding_table(
        self,
        table: str,
        columns: List[str],
        schema: str = 'public',
        embedding_column: str = 'embedding',
        document_column: str = 'document',
        metadata_column: str = 'metadata',
        id_column: str = 'id',
        dimension: int = 768,
        use_jsonb: bool = False,
        drop_columns: bool = False,
        **kwargs
    ):
        """
        Create an embedding column and vectorize Table information.
        """
        tablename = f'{schema}.{table}'
        cols = ', '.join(columns)
        _qry = f'SELECT {cols} FROM {tablename};'
        # Generate a sample embedding to determine its dimension
        sample_vector = self._embed_.embedding.embed_query("sample text")
        vector_dim = len(sample_vector)
        # Compare it with the expected dimension
        if vector_dim != dimension:
            raise ValueError(
                f"Expected embedding dimension {self.dimension}, but got {vector_dim}"
            )
        async with self._connection.begin() as conn:
            result = await conn.execute(
                sqlalchemy.text(_qry)
            )
            rows = result.fetchall()
            # Concatenate column names and values to form an input string:
            # 'store_name: BestBuy, location_code: 123456, ...'
            # data = [
            #     ', '.join([f"{col}: {row[col]}" for col in columns])
            #     for row in rows
            # ]
            # if drop columns, then first remove the existing columns:
            if drop_columns:
                for column in (document_column, embedding_column, metadata_column):
                    await conn.execute(
                        sqlalchemy.text(
                            f'ALTER TABLE {tablename} DROP COLUMN IF EXISTS {column};'
                        )
                    )
            # Create a new column for embeddings
            if use_jsonb:
                await conn.execute(
                    sqlalchemy.text(
                        f'ALTER TABLE {tablename} ADD COLUMN IF NOT EXISTS {embedding_column} JSONB;'  # pylint: disable=C0301
                    )
                )
            else:
                # Use Embedding pgvector type:
                await conn.execute(
                    sqlalchemy.text(
                        f'ALTER TABLE {tablename} ADD COLUMN IF NOT EXISTS {embedding_column} vector({dimension});'  # pylint: disable=C0301
                    )
                )
                # Create a Index for vector:
                # TODO: define index algorithm and options.
                await conn.execute(
                    sqlalchemy.text(
                        f"CREATE INDEX IF NOT EXISTS idx_{schema}_{table}_embeddings ON {tablename} USING hnsw ({embedding_column} vector_l2_ops);"  # pylint: disable=C0301
                    )
                )
                # And also, an index IVFLAT:
                await conn.execute(
                    sqlalchemy.text(
                        f"CREATE INDEX IF NOT EXISTS idx_{schema}_{table}_ivflat ON {tablename} USING ivfflat ({embedding_column} vector_cosine_ops);"  # pylint: disable=C0301
                    )
                )
            # Then, create the info column and id column (if required):
            # Text info Column (content)
            await conn.execute(
                sqlalchemy.text(
                    f'ALTER TABLE {tablename} ADD COLUMN IF NOT EXISTS {document_column} TEXT;'
                )
            )
            # ID Column (if required)
            await conn.execute(
                sqlalchemy.text(
                    f'ALTER TABLE {tablename} ADD COLUMN IF NOT EXISTS {id_column} varchar;'
                )
            )
            # Metadata Column (JSONB):
            await conn.execute(
                sqlalchemy.text(
                    f'ALTER TABLE {tablename} ADD COLUMN IF NOT EXISTS {metadata_column} jsonb;'
                )
            )
            # And ID Column:
            await conn.execute(
                sqlalchemy.text(
                    f'ALTER TABLE {tablename} ADD COLUMN IF NOT EXISTS {id_column} varchar;'
                )
            )
            for row in rows:
                _id = getattr(row, id_column)
                metadata = {col: getattr(row, col) for col in columns}
                data = " ".join([f"{col}: {metadata[col]}" for col in columns])
                # Get the vector information from data:
                vector = self._embed_.embedding.embed_query(data)
                vector_str = "[" + ",".join(str(v) for v in vector) + "]"
                await conn.execute(
                sqlalchemy.text(f"""
                    UPDATE {tablename}
                    SET {embedding_column} = :vector, {document_column} = :info, {metadata_column} = :metadata
                    WHERE {id_column} = :id
                """),
                {"vector": vector_str, "id": _id, "info": data, "metadata": json_encoder(metadata)}
            )
        print("✅ Updated Table embeddings.")



from .abstract import AbstractBot

class BasicBot(AbstractBot):
    """Represents an BasicBot in Navigator.

        Each BasicBot has a name, a role, a goal, a backstory,
        and an optional language model (llm).
    """
    pass
